from pathlib import Path
import sys
import os
import platform
import getpass
import traceback
from multiprocessing import Pool
from typing import Dict, List, Tuple, Type
from unittest import TestCase
import uuid
from datetime import datetime, timezone
from gempyp.config.baseConfig import AbstarctBaseConfig
# from gempyp.engine.baseTemplate import TestcaseReporter as Base
from gempyp.engine.testData import TestData
from gempyp.libs.enums.status import status
from gempyp.libs.enums.run_types import RunTypes
from gempyp.reporter.reportGenerator import TemplateData
from gempyp.libs import common
from gempyp.engine.runner import testcaseRunner
from gempyp.config import DefaultSettings
import logging
from gempyp.libs.logConfig import my_custom_logger, LoggingConfig
from gempyp.engine import dataUpload
from gempyp.pyprest.pypRest import PypRest
import smtplib
from gempyp.dv.dvRunner import DvRunner
from gempyp.jira.jiraIntegration import jiraIntegration
from gempyp.jira.azureIntegration import azureIntegration
from multiprocessing import Process, Pipe
from gempyp.libs.gem_s3_common import upload_to_s3, create_s3_link, uploadToS3
from gempyp.libs.common import *
import re
from importlib.metadata import version
import pandas as pd


def executorFactory(data: Dict, conn=None, custom_logger=None) -> Tuple[List, Dict]:
    """
    calls the differnt executors method based on testcase type e.g. gempyp,pyprest,dvm
    Takes single testcase data as input
    """
    logging.info("--------- In Executor Factory ----------\n")
    if custom_logger == None:
        testcase_name = data['config_data'].get('NAME', None)
        testcase_name = re.sub('[^A-Za-z0-9]+', '_', testcase_name)
        log_path = os.path.join(os.environ.get('TESTCASE_LOG_FOLDER'), testcase_name + '_'
                                + os.environ.get('unique_id') + '.txt')  # replacing log with txt for UI compatibility
        if len(log_path) > 240:
            log_path = os.path.join(os.environ.get(
                'TESTCASE_LOG_FOLDER'), str(uuid.uuid4()) + '.txt')
        custom_logger = my_custom_logger(log_path)
        LoggingConfig(log_path)
    data['config_data']['LOGGER'] = custom_logger
    if 'log_path' not in data['config_data']:
        data['config_data']['LOG_PATH'] = log_path

    engine_control = {
        "pyprest": {"class": PypRest, "classParam": data, "function": "restEngine"},
        "dv": {"class": DvRunner, "classParam": data, "function": "dvEngine"},
        "gempyp": {"function": testcaseRunner, "functionParam": data}
    }
    _type = data.get("config_data").get("TYPE", "GEMPYP") if data.get(
        "config_data").get("TYPE", None) else "GEMPYP"
    dv = ["data validator","dv","datavalidator","dvalidator"]
    pyprest = ["pyprest","prest","pr"]
    gempyp = ["gempyp","gpyp","gp"]
    if _type in dv:
        _type = "dv"
    elif _type in pyprest:
        _type = "pyprest"
    elif _type in gempyp:
        _type = "gempyp"

    _type_dict = engine_control[_type.lower()]
    custom_logger.info(f"Starting {_type} testcase")

    if _type_dict.get("class", None):
        # need to make it generic for functionParam too
        data = getattr(_type_dict["class"](
            _type_dict['classParam']), _type_dict['function'])()
    else:
        # we need to dissolve this else condition too somehow
        data = _type_dict["function"](_type_dict["functionParam"])
    if conn == None:
        return data
    else:
        conn.send([data])
        conn.close()


class Engine:
    def __init__(self, params_config):
        """
        constructor used to  call run method
        takes config as input
        """
        # logging.basicConfig()
        # logging.root.setLevel(logging.DEBUG)
        sorted_config = self.sortTestcase(getattr(params_config, "_CONFIG"))
        setattr(params_config, "_CONFIG", sorted_config)
        self.run(params_config)

    def run(self, params_config: Type[AbstarctBaseConfig]):
        """
        main method to call other methods that are required for report generation
        takes config as input
        """
        logging.info("Engine Started")
        # initialize the data class

        self.DATA = TestData()
        # get the env for the engine Runner
        # self.ENV = os.getenv("appenv", "BETA").upper()
        # initial SETUP
        self.setUP(params_config)
        self.makeSuiteDetails()
        # handle case of dependency
        validateZeroTestcases(self.CONFIG.getTestcaseLength())
        runBaseUrls(self.jewel_user, self.base_url, self.username,
                    self.bridgetoken)  # retrying to run Base Urls
        self.DATA.validateSrunidInDB(
            self.jewel_user, self.s_run_id, self.username, self.bridgetoken)
        self.makeOutputFolder()
        self.start()

        if (self.jewel_user and DefaultSettings.apiSuccess):
            self.DATA.retryUploadSuiteData(self.bridgetoken, self.username)
        jewel, failed_Utestcases = self.DATA.retryUploadTestcases(
            self.s_run_id, self.bridgetoken, self.username, self.ouput_folder)
        self.updateSuiteData()
        unuploaded_path = None
        if dataUpload.suite_uploaded and DefaultSettings.apiSuccess:
            dataUpload.sendSuiteData(self.DATA.toSuiteJson(
            ), self.bridgetoken, self.username, mode="PUT")

            if self.skip_jira == 0:
                jira_id = jiraIntegration(self.s_run_id, self.jira_email, self.jira_access_token, self.jira_project_id, self.project_env, self.jira_workflow,
                                        self.jira_title, self.bridgetoken, self.username, self.report_name)  # adding title  ######################### post 1.0.4
                if jira_id is not None:
                    self.DATA.suite_detail.at[0, "meta_data"].append(
                        {"Jira_id": jira_id})
                    
            if self.skip_azure == 0:
                azure_id = azureIntegration(self.s_run_id, self.azure_assigned_to, self.azure_pat, self.azure_fields, self.azure_project, self.project_env, self.azure_workflow,
                                        self.azure_testcase_flag, self.azure_title, self.bridgetoken, self.username, self.report_name, self.azure_organization)  # adding title  ######################### post 1.0.4
                if azure_id is not None:
                    self.DATA.suite_detail.at[0, "meta_data"].append(
                        {"Azure_id": azure_id})
            # # dataUpload.sendSuiteData(self.DATA.toSuiteJson(), self.PARAMS["BRIDGE_TOKEN"], self.PARAMS["USERNAME"], mode="PUT")
        else:
            unuploaded_path = self.DATA.WriteSuiteFile(
                self.base_url, self.ouput_folder, self.username, self.bridgetoken)

        if ("EMAIL_TO" in self.PARAMS.keys()):
            sendMail(self.s_run_id, self.mail, self.bridgetoken, self.username)

        self.repJson, final_report_path = TemplateData().makeSuiteReport(self.DATA.getJSONData(),
                                                    self.testcase_data, self.ouput_folder, self.jewel_user)
        TemplateData().repSummary(self.repJson, jewel, unuploaded_path, self.testcase_log_folder,
                                self.complete_logs, self.bridgetoken, self.username, self.suite_log_file, final_report_path)

    def makeOutputFolder(self):
        """
        make outputFolder for report named as gempyp_reports in user home directory if not given by the user and makes log fplder for log files
        if given by user than set user given path for reports file   
        """

        logging.info("---------- Making output folders -------------")
        report_folder_name = f"{self.project_name}_{self.project_env}"
        if self.report_name:
            report_name = "_".join(self.report_name.split())
            report_folder_name = report_folder_name + f"_{report_name}"
        date = datetime.now().strftime("%Y_%b_%d_%H%M%S_%f")
        report_folder_name = report_folder_name + f"_{date}"
        try:
            if "REPORT_LOCATION" in self.PARAMS and self.PARAMS["REPORT_LOCATION"]:
                self.ouput_folder = os.path.join(
                    self.PARAMS["REPORT_LOCATION"], report_folder_name
                )
            else:
                home = str(Path.home())
                self.ouput_folder = os.path.join(
                    home, "gempyp_reports", report_folder_name
                )
        except Exception:
            temp = str(tempfile.gettempdir())
            self.ouput_folder = os.path.join(
                temp, "gempyp_reports", report_folder_name
            )
        os.makedirs(self.ouput_folder)
        self.testcase_folder = os.path.join(self.ouput_folder, "testcases")
        os.makedirs(self.testcase_folder)
        self.testcase_log_folder = os.path.join(self.ouput_folder, "logs")
        os.environ['TESTCASE_LOG_FOLDER'] = self.testcase_log_folder
        os.makedirs(self.testcase_log_folder)
        self.complete_logs = os.path.join(
            self.testcase_log_folder, self.s_run_id + '.txt')

    def verify(self, value):
        if (re.search(r'[^a-zA-Z0-9_ .-]', value)):
            logging.info(
                "Some Error From the Client Side. May be s_run_id, project_name or ENV is not in a correct format")
            sys.exit()
        else:
            return value

    def setUP(self, config: Type[AbstarctBaseConfig]):
        """

        assigning values to some attributes which will be used in method makeSuiteDetails

        """

        self.PARAMS = config.getSuiteConfig()
        self.suite_log_file = config.getLogFilePath()
        self.ENV = os.getenv("appenv", "BETA").upper()

        #checking if url is present in file and calling get api
        self.CONFIG = config

        self.testcase_data = {}

        self.total_runable_testcase = config.total_yflag_testcase

        self.machine = platform.node()
        # creating job_name variable to set job name from suite tags reason:- facing issue on lambda side
        self.job_name = self.PARAMS.get("JEWEL_JOB",None)

        self.user = self.PARAMS.get("JEWEL_USER", getpass.getuser())
        self.username = self.PARAMS.get("JEWEL_USER", None)
        self.bridgetoken = self.PARAMS.get("JEWEL_BRIDGE_TOKEN", None)
        self.base_url = self.PARAMS.get("ENTER_POINT", None)
        self.autoKill = self.PARAMS.get("AUTO_KILL", 'y')
        if self.autoKill.lower() not in ['y', 'n', "true", "false"]:
            self.autoKill = 'y'
            logging.info(
                "Auto kill value given by user is not accurate, keeping it as Y")
        # INVOKEUSER can be set as environment variable from anywhere.
        self.invoke_user = os.getenv("INVOKEUSER", self.user)
        self.current_dir = os.getcwd()

        self.platform = platform.system()

        self.start_time = datetime.now(timezone.utc)
        self.skip_jira = 0
        self.skip_azure = 0
        try:
            self.jira_email = self.PARAMS.get("JIRA_EMAIL", None)
            self.jira_access_token = self.PARAMS.get("JIRA_ACCESS_TOKEN", None)
            self.jira_project_id = self.PARAMS.get("JIRA_PROJECT_ID", None)
            self.jira_workflow = self.PARAMS.get("JIRA_WORKFLOW", None)
            # adding title  ######################### post 1.0.4
            self.jira_title = self.PARAMS.get("JIRA_TITLE", None)
            if self.jira_access_token is None and self.jira_email is None:
                self.skip_jira = 1

            self.azure_pat = self.PARAMS.get("AZURE_PAT", None)
            self.azure_project = self.PARAMS.get("AZURE_PROJECT", None)
            self.azure_testcase_flag = self.PARAMS.get("AZURE_TESTCASE_FLAG", "N")
            self.azure_assigned_to = self.PARAMS.get("AZURE_ASSIGNED_TO", None)
            self.azure_workflow = self.PARAMS.get("AZURE_WORKFLOW", None)
            self.azure_fields = self.PARAMS.get("AZURE_FIELDS", None)
            self.azure_title = self.PARAMS.get("AZURE_TITLE", None)
            self.azure_organization = self.PARAMS.get("AZURE_ORGANIZATION", None)

            if self.azure_testcase_flag.upper().strip() != 'Y' and self.azure_testcase_flag.upper().strip() != 'N':
                logging.info("User entered wrong azure_testcase_flag!")
                logging.info("Thus disabling it by default.")
                self.azure_testcase_flag = "N"

            try:
                if self.azure_fields:
                    self.azure_fields = json.loads(self.azure_fields)
            except Exception as e:
                logging.error("User has entered wrong data in Azure_fields!")
                self.skip_azure = 1

            if self.azure_pat is None or self.azure_project is None or self.azure_organization is None:
                self.skip_azure = 1
                logging.info("Not enough data for azure integration! Thus skipping it")

        except Exception as e:
            logging.error("Exception Occurred While Fetching Jira/ Azure Details")

        mail_items = {"to": "EMAIL_TO", "cc": "EMAIL_CC", "bcc": "EMAIL_BCC"}
        self.mail = {key: common.parseMails(self.PARAMS.get(
            value, None)) for key, value in mail_items.items()}

        self.project_name = self.verify(self.PARAMS["PROJECT_NAME"])
        self.project_env = self.verify(self.PARAMS["ENVIRONMENT"])
        self.report_name = strValidation(self.PARAMS.get("REPORT_NAME"))
        self.unique_id = self.PARAMS["UNIQUE_ID"]
        self.user_suite_variables = self.PARAMS.get("SUITE_VARS", {})
        self.user_global_variables = self.PARAMS.get("GLOBAL_VARIABLES", {})
        self.jewel_run = False
        self.jewel_user = False
        self.s3_url = ""
        if self.bridgetoken and self.username:
            self.user_suite_variables["bridge_token"] = self.bridgetoken
            self.user_suite_variables["username"] = self.username
            self.jewel_user = True

        runBaseUrls(self.jewel_user, self.base_url, self.username,
                    self.bridgetoken)  # Run base Urls
        if self.jewel_user:
            # trying first run of base url api in case of api failure

            if self.PARAMS.get("S_ID", None):
                self.jewel_run = True
            else:
                try:
                    if DefaultSettings.apiSuccess:
                        self.s3_url = upload_to_s3(DefaultSettings.urls["data"]["bucket-file-upload-api"],
                                                bridge_token=self.bridgetoken, username=self.username, file=self.PARAMS["config"])[0]["Url"]
                        logging.info("S3 Url: " + str(self.s3_url))
                    else:
                        self.s3_url = self.PARAMS["config"]
                except Exception as e:
                    logging.info(e)

    def makeSuiteDetails(self):
        """
        making suiteDetails dictionary and assign it to DATA.suiteDetail 
        """
        if "S_RUN_ID" in self.PARAMS:
            self.s_run_id = self.verify(self.PARAMS["S_RUN_ID"])
        else:
            if not self.unique_id:
                self.unique_id = uuid.uuid4()
            self.s_run_id = f"{self.project_name}_{self.project_env}_{self.unique_id}"
            self.s_run_id = self.s_run_id.upper()
        logging.info("S_RUN_ID: {}".format(self.s_run_id))
        package_name = "gempyp"
        try:
            version1 = version(package_name)  # Get version from installed package metadata
        except Exception:
            version1 = None

        # If package is not installed, try to get version from pyproject.toml
        if version1 is None:
            version1 = self.get_version_from_toml_file()
        if(platform.system().upper()=="WINDOWS"):
            osName=platform.system().upper()+" "+platform.version().split(".")[0]
        else:
            osName=platform.system().upper()
        suite_details = {
            "s_run_id": self.s_run_id,
            "s_start_time": self.start_time,
            "s_end_time": None,
            "s_id": self.PARAMS.get("S_ID", "test_id"),
            "status": status.EXE.name,
            "project_name": self.project_name,
            "report_name": self.report_name,  # earlier it was report info
            "user": self.user,
            "env": self.project_env,
            "machine": self.machine,
            "framework_version":version1,
            "os": osName,
            "meta_data": [],
            "expected_testcases": self.total_runable_testcase,
            "testcase_info": None,
            "framework_name": "GEMPYP",
            "autoKill": self.autoKill.lower()
        }
        # self.DATA.suite_detail = self.DATA.suite_detail.append(
        #     suite_details, ignore_index=True
        # )
        self.DATA.suite_detail = pd.concat([self.DATA.suite_detail, pd.DataFrame([suite_details])], ignore_index=True)

    def get_version_from_toml_file(self):
        try:
            import toml
            with open('pyproject.toml', 'r') as toml_file:
                content = toml.load(toml_file)
                return content['tool']['poetry']['version']
        except FileNotFoundError:
            print("pyproject.toml file not found. Defaulting to version 0.0.0")
            return "0.0.0"
        except Exception as e:
            print(f"An error occurred while reading pyproject.toml: {e}, defaulting to version 0.0.0")
            return "0.0.0"

    
    # def get_version_from_setup(self):
    #     try:
    #         with open('setup.py', 'r') as setup_file:
    #             content = setup_file.read()
    #             # Extract version using regular expression
    #             match = re.search(r"version=['\"](.*?)['\"]", content)
    #             if match:
    #                 return match.group(1)
    #     except FileNotFoundError:
    #         pass  # Handle the case when setup.py is not found
    #     except Exception as e:
    #         print(f"An error occurred while reading setup.py: {e}")

    #     return None

    def start(self):
        """
        check the mode and start the testcases accordingly e.g.optimize,parallel
        check the mode and start the testcases accordingly e.g.optimize,parallel
        """
        try:
            if self.PARAMS["MODE"].upper() == "SEQUENCE":
                self.startSequence()
            elif self.PARAMS["MODE"].upper() == "OPTIMIZE" or self.PARAMS.get("MODE", None) is None:
                self.startParallel()
            else:
                raise TypeError("mode can only be sequence or optimize")

        except Exception as e:
            logging.error(traceback.format_exc())
            try:
                self.DATA.suite_detail.at[0, "meta_data"].append(
                    {"REASON OF FAILURE": str(e)})
                self.updateSuiteData()
            except Exception as err:
                logging.error(traceback.format_exc())
                logging.info(err)
            if DefaultSettings.apiSuccess:
                dataUpload.sendSuiteData(
                    (self.DATA.toSuiteJson()), self.bridgetoken, self.username)
            else:
                dataUpload.suite_data.append(self.DATA.toSuiteJson())
            # need to add reason of failure of the suite in misc

    def updateSuiteData(self):
        """
        updates the suiteData after all the runs have been executed
        """

        # get the status count of the status
        status_dict = self.DATA.testcase_details["status"].value_counts(
        ).to_dict()
        total = sum(status_dict.values())
        status_dict["TOTAL"] = total
        unsorted_dict = status_dict
        sorted_dict = self.totalOrder(unsorted_dict)
        status_dict = sorted_dict
        # status_dict = dict( sorted(status_dict.items(), key=lambda x: x[0].lower(), reverse=True) )
        Suite_status = status.FAIL.name

        # based on the status priority
        for s in status:
            if status_dict.get(s.name, 0) > 0:
                Suite_status = s.name
                break
        stop_time = (
            self.DATA.testcase_details["end_time"].max()
        )
        self.DATA.suite_detail.at[0, "status"] = Suite_status
        self.DATA.suite_detail.at[0, "s_end_time"] = stop_time
        self.DATA.suite_detail.at[0, "testcase_info"] = status_dict
        if self.jewel_run is True:
            self.DATA.suite_detail.at[0, "meta_data"].append(
                {"S_ID": self.PARAMS["S_ID"]})
        else:
            self.DATA.suite_detail.at[0, "meta_data"].append(
                {"CONFIG_S3_URL": self.s3_url})

    def startSequence(self):
        """
        start calling executoryFactory() for each testcase one by one according to their dependency
        at last of each testcase calls the update_df() 
        """
        response = False
        for testcases in self.getDependency(self.CONFIG.getTestcaseConfig()):
            if response:
                break
            for testcase in testcases:
                passedDependency = self.isDependencyPassed(testcase)
                product_type = {'dv': "GEMPYP-DV",
                                    "pyprest": "GEMPYP-PR", "gempyp": "GEMPYP"}
                dependency_error = {
                    "message": "dependency failed",
                    "testcase": testcase["NAME"],
                    "category": testcase.get("CATEGORY", None),
                    "product_type": product_type.get(testcase.get("TYPE", None), testcase.get("TYPE", None)),
                }
                if passedDependency == 'true':
                    data = self.getTestcaseData(testcase['NAME'])
                    testcase_name = data['config_data'].get('NAME', None)
                    testcase_name = re.sub('[^A-Za-z0-9]+', '_', testcase_name)
                    log_path = os.path.join(self.testcase_log_folder,
                                            testcase_name+'_'+self.CONFIG.getSuiteConfig()['UNIQUE_ID'] + '.txt')  # ## replacing log with txt for UI compatibility
                    if len(log_path) > 240:
                        log_path = os.path.join(os.environ.get(
                            'TESTCASE_LOG_FOLDER'), str(uuid.uuid4()) + '.txt')
                    custom_logger = my_custom_logger(log_path)
                    data['config_data']['log_path'] = log_path
                    conn = None
                    output, error = executorFactory(data, conn, custom_logger)
                    if output is not None and output[0]["GLOBAL_VARIABLES"].get("UPDATED_GLOBAL_VARS", None) is not None:
                        key_list = output[0]["GLOBAL_VARIABLES"]["UPDATED_GLOBAL_VARS"]
                        for key in key_list:
                            logging.info(
                                " Updating global variables values after testcase execution -- {k}".format(k=key))
                            self.user_global_variables[key] = output[0]["GLOBAL_VARIABLES"][key]
                    response = self.update_df(output, error)
                    if response:
                        break
                elif passedDependency == 'err':
                    dependency_error["status"] = status.ERR.name
                    self.update_df(None, dependency_error)
                elif passedDependency == 'fail':
                    dependency_error["status"] = status.ERR.name
                    self.update_df(None, dependency_error)

    def startParallel(self):
        """
        start calling executorFactory for testcases in parallel according to their drependency 
        at last of each testcase calls the update_df()
        """
        pool = None
        try:

            threads = self.PARAMS.get("THREADS", DefaultSettings.THREADS)
            try:
                threads = int(threads)
            except:
                threads = DefaultSettings.THREADS
            # pool = Pool(threads)
            response = False
            for testcases in self.getDependency(self.CONFIG.getTestcaseConfig()):
                if response:
                    break

                # create a list to keep connections
                if len(testcases) == 0:
                    raise Exception("No testcase to run")
                pool_list = []
                for testcase in testcases:
                    # only append testcases whose dependency are passed otherwise just update the databasee
                    passedDependency = self.isDependencyPassed(testcase)
                    product_type = {'dv': "GEMPYP-DV",
                                        "pyprest": "GEMPYP-PR", "gempyp": "GEMPYP"}
                    dependency_error = {
                        "message": "dependency failed",
                        "testcase": testcase["NAME"],
                        "category": testcase.get("CATEGORY", None),
                        "product_type": product_type.get(testcase.get("TYPE", None), testcase.get("TYPE", None)),
                        "status": status.ERR.name
                    }
                    if passedDependency == 'true':
                        pool_list.append(
                            self.getTestcaseData(testcase.get("NAME")))
                    elif passedDependency == 'err':
                        dependency_error["status"] = status.ERR.name
                        self.update_df(None, dependency_error)
                    elif passedDependency == 'fail':
                        dependency_error["status"] = status.FAIL.name
                        self.update_df(None, dependency_error)


                if len(pool_list) == 0:
                    continue

                # obj = Semaphore(8)
                splitedSize = threads
                a_splited = [pool_list[x:x+splitedSize]
                            for x in range(0, len(pool_list), splitedSize)]
                for chunk_list in a_splited:
                    if response:
                        break
                    processes = []
                    parent_connections = []
                    for testcase in chunk_list:
                        testcase["default_urls"] = DefaultSettings.urls
                        parent_conn, child_conn = Pipe()
                        parent_connections.append(parent_conn)

                # create the process, pass instance and connection
                        custom = None
                        process = Process(target=executorFactory, args=(
                            testcase, child_conn, custom))
                        processes.append(process)
                # runs the testcase in parallel here
                    instances_total = []

                    # print(a_splited)
                    # for i in a_splited:
                    for process in processes:
                        process.start()
                    for parent_connection in parent_connections:

                        instances_total.append(parent_connection.recv()[0])
                        if len(instances_total) > 0:
                            row = instances_total[-1]
                            if not row or len(row) < 2:
                                raise Exception(
                                    "Some error occurred while running the testcases"
                                )
                            output = row[0]

                            error = row[1]
                            if error:
                                logging.error(
                                    f"Error occurred while executing the testcase: {error['testcase']}"
                                )
                                logging.error(f"message: {error['message']}")

                            if output is not None and output[0]["GLOBAL_VARIABLES"].get("UPDATED_GLOBAL_VARS", None) is not None:
                                key_list = output[0]["GLOBAL_VARIABLES"]["UPDATED_GLOBAL_VARS"]
                                for key in key_list:
                                    logging.info(
                                        " Updating global variables values after testcase execution -- {k}".format(k=key))
                                    self.user_global_variables[key] = output[0]["GLOBAL_VARIABLES"][key]

                            response = self.update_df(output, error)


                    for process in processes:
                        process.join()
        except Exception:
            logging.error(traceback.format_exc())

    def update_df(self, output: List, error: Dict):
        """
        updates the testcase data in the dataframes of testData.py
        also upload testcasedata to db

        """
        try:
            if error:
                output = self.getErrorTestcase(
                    error["message"],
                    error["testcase"],
                    error.get("category"),
                    error.get("product_type"),
                    error.get('log_path', None),
                    error.get('status', status.ERR)
                )
                output = [output]
            if 'json_data' in output[0]:
                unsorted_dict = output[0]['json_data']['meta_data'][2]
                sorted_dict = self.totalOrder(unsorted_dict)
                output[0]['json_data']['meta_data'][2] = sorted_dict

            output[0]['testcase_dict']['run_type'], output[0]['testcase_dict']['run_mode'], output[0][
                'testcase_dict']['job_name'], output[0]['testcase_dict']['job_runid'] = self.set_run_type_mode()
            for i in output:
                if 'json_data' in i:
                    i["testcase_dict"]["steps"] = i["json_data"]["steps"]
                else:
                    i["testcase_dict"]["steps"] = self.build_err_step_case()

                testcase_dict = i["testcase_dict"]
                try:
                    """ update suite vars here from testcase_dict["suite_variables"] append it in the suite vars of _config"""

                    self.user_suite_variables.update(
                        i.get("suite_variables", {}))

                    self.testcase_data[testcase_dict.get(
                        "tc_run_id")] = i["json_data"]
                except Exception as e:
                    logging.error(e)
                # self.DATA.testcase_details = self.DATA.testcase_details.append(
                #     testcase_dict, ignore_index=True
                # )
                self.DATA.testcase_details = pd.concat([self.DATA.testcase_details, pd.DataFrame([testcase_dict])], ignore_index=True
                )
                self.updateTestcaseMiscData(
                    i["misc"], tc_run_id=testcase_dict.get("tc_run_id")
                )
                if (self.jewel_user and dataUpload.suite_uploaded):
                    response = dataUpload.sendTestcaseData((self.DATA.totestcaseJson(testcase_dict.get(
                        "tc_run_id").upper(), self.s_run_id)), self.bridgetoken, self.username)
                    return True if response == 412 else False

                else:
                    dataUpload.not_uploaded.append((self.DATA.totestcaseJson(
                        testcase_dict.get("tc_run_id").upper(), self.s_run_id)))

        except Exception as e:
            traceback.print_exc()
            logging.error("in update_df: {e}".format(e=e))

    def build_err_step_case(self):
        step = [{'Step Name': 'Starting Test', 'Step Description':
                'Either the testcase is inappropriate or some error occurred while executing the test. Please recheck', 'status': 'ERR'}]

        return step

    def raise_exception(self, name):
        raise ValueError("{} does not exist".format(name))

    def set_run_type_mode(self):
        try:
            run_type=run_mode=job_name=job_runid=None
            mappings = {('JEWEL'): { 
                'run_type': 'Scheduled', 
                'run_mode': 'JEWEL_SCHEDULER',
                'job_name': lambda: os.environ.get('JEWEL_JOB',None), 
                'job_runid': 'NONE' 
            },
                ('SCHEDULER_TOOL'): {
                'run_type': 'Scheduled',
                'run_mode': lambda: os.environ.get('SCHEDULER_TOOL'),
                'job_name': lambda: os.environ.get('SCHEDULER_JOB') if os.environ.get('SCHEDULER_JOB', None) else self.raise_exception("Schedular job"),
                'job_runid': lambda: os.environ.get('SCHEDULER_RUNNUM') if os.environ.get('SCHEDULER_RUNNUM', None) else self.raise_exception("Schedular rennum")
            },
                ('CI_CD_CT_TOOL'): {
                'run_type': 'CI-CD-CT',
                'run_mode': lambda: os.environ.get('CI_CD_CT_TOOL') if os.environ.get('CI_CD_CT_TOOL', None) else self.raise_exception("CI_CD_CT_TOOL"),
                'job_name': lambda: os.environ.get('CI_CD_CT_JOB') if os.environ.get('CI_CD_CT_JOB', None) else self.raise_exception('CI_CD_CT_JOB'),
                'job_runid': lambda: os.environ.get('CI_CD_CT_RUNNUM') if os.environ.get('CI_CD_CT_RUNNUM', None) else self.raise_exception('CI_CD_CT_RUNNUM')
            },
                ('AUTO_JOB_NAME'): {
                'run_type': 'Scheduled',
                'run_mode': 'Autosys',
                'job_name': lambda: os.environ.get('AUTOSERV') + '.' + os.environ.get('AUTO_JOB_NAME') if os.environ.get('AUTOSERV', None) and os.environ.get('AUTO_JOB_NAME', None) else self.raise_exception("AUTOSERV or AUTO_JOB_NAME"),
                'job_runid': lambda: os.environ.get('AUTO_JOBID') if os.environ.get('AUTO_JOBID', None) else self.raise_exception("AUTO_JOBID")
            },
                ('JENKINS_URL'): {
                'run_type': lambda: 'Scheduled' if os.environ.get('BUILD_CAUSE') == 'TIMERTRIGGER' else 'CI-CD-CT' if os.environ.get('BUILD_CAUSE') == 'SCMTRIGGER' else 'On Demand',
                'run_mode': 'Jenkins',
                'job_name': lambda: os.environ.get('JOB_DISPLAY_URL') if os.environ.get('JOB_DISPLAY_URL', None) else self.raise_exception("JOB_DISPLAY_URL"),
                'job_runid': lambda: os.environ.get('BUILD_URL') if os.environ.get('BUILD_URL', None) else self.raise_exception("BUILD_URL")}
            }
            # Try to match the environment variables to one of the defined mappings
            for env_vars, mapping in mappings.items():
                if (env_vars in os.environ):
                    # Evaluate the values that require computation (i.e. lambda functions)
                    run_type = mapping['run_type']() if callable(
                        mapping['run_type']) else mapping['run_type']
                    run_mode = mapping['run_mode']() if callable(
                        mapping['run_mode']) else mapping['run_mode']
                    job_name = mapping['job_name']() if callable(
                        mapping['job_name']) else mapping['job_name']
                    job_runid = mapping['job_runid']() if callable(
                        mapping['job_runid']) else mapping['job_runid']
                    break
                else:
                    # If no mapping is found, set default values
                    run_type = 'On Demand'
                    # run_mode = os.name
                    run_mode = platform.uname().system
                    job_name = 'NONE'
                    job_runid = 'NONE'
        except ValueError as e:
            logging.error(e)
        return run_type, run_mode, job_name, job_runid

    def getErrorTestcase(
        self,
        message: str,
        testcase_name: str,
        category: str = None,
        product_type: str = None,
        log_path: str = None,
        status: str = None
    ) -> Dict:
        """
        store the data of failed testcase and return it as a dict to update_df
        take message for error as input
        """

        result = {}
        testcase_dict = {}
        misc = {}
        if not self.unique_id:
            self.unique_id = uuid.uuid4()
        # testcase should not be testcase + s_run_id
        tc_run_id = f"{testcase_name}_{self.unique_id}"

        tc_run_id = tc_run_id.upper()
        testcase_dict["tc_run_id"] = tc_run_id
        testcase_dict["status"] = status
        testcase_dict["start_time"] = datetime.now(timezone.utc)
        testcase_dict["end_time"] = datetime.now(timezone.utc)
        testcase_dict["name"] = testcase_name
        testcase_dict["ignore"] = False
        if category:
            testcase_dict["category"] = category
        s3_log_file_url = log_path
        if self.jewel_user:
            try:
                s3_log_file_url= uploadToS3(DefaultSettings.urls["data"].get("pre-signed",None), bridge_token=self.bridgetoken, username=self.username, file=log_path,tag="protected",folder="logs",s_run_id=self.s_run_id)[0]
                # s3_log_file_url = f'<a href="{s3_log_file_url}" target=_blank>view</a>'
            except Exception as e:
                logging.info(e)
        testcase_dict["log_file"] = log_path
        testcase_dict["result_file"] = None
        testcase_dict["base_user"] = getpass.getuser()
        testcase_dict["invoke_user"] = self.invoke_user
        testcase_dict["machine"] = self.machine
        # testcase_dict["response_time"]="{0:.{1}f} sec(s)".format((testcase_dict["end_time"]-testcase_dict["start_time"]).total_seconds(),2)
        if product_type:
            testcase_dict["product_type"] = product_type
        result["testcase_dict"] = testcase_dict
        misc["REASON OF FAILURE"] = message
        result["misc"] = misc
        result["misc"]["log_file"] = s3_log_file_url
        # result["json_data"]["meta_data"][2]["TOTAL"] = total   # we can not get dummy data because here testcase does not exist
        return result

    def updateTestcaseMiscData(self, misc: Dict, tc_run_id: str):
        """
        updates the misc data for the testcases in testData.py
        accept miscellaneous rows and tc_run_id as parameters
        """
        misc_list = []

        for misc_data in misc:
            temp = {}
            # storing all the key in upper so that no duplicate data is stored
            temp["key"] = misc_data.upper()
            temp["value"] = misc[misc_data]
            temp["run_id"] = tc_run_id
            temp["table_type"] = "TESTCASE"
            misc_list.append(temp)

        # self.DATA.misc_details = self.DATA.misc_details.append(
        #     misc_list, ignore_index=True
        # )
        self.DATA.misc_details = pd.concat([self.DATA.misc_details, pd.DataFrame([misc_list])], ignore_index=True)

    def getTestcaseData(self, testcase: str) -> Dict:
        """
        taking argument as the testcase name and  return dictionary containing information about testCase
        """
        data = {}
        list_subtestcases = []
        data["config_data"] = self.CONFIG.getTestcaseData(testcase)
        if ("SUBTESTCASES" in data["config_data"].keys()):
            for key1 in data["config_data"]["SUBTESTCASES"].split(","):
                list_subtestcases.append(self.CONFIG.getSubTestcaseData(key1))
            data["config_data"]["SUBTESTCASES_DATA"] = list_subtestcases
        data["PROJECT_NAME"] = self.project_name
        data["ENVIRONMENT"] = self.project_env
        if (self.project_env.upper() in self.PARAMS.keys()):
            data[self.project_env.upper()] = self.PARAMS[self.project_env.upper()]
        data["S_RUN_ID"] = self.s_run_id
        # data["USER"] = self.user
        data["MACHINE"] = self.machine
        data["REPORT_LOCATION"] = self.testcase_folder
        data["SUITE_VARS"] = self.user_suite_variables
        data["INVOKE_USER"] = self.invoke_user
        data["USER"] = self.user
        data['GLOBAL_VARIABLES'] = self.user_global_variables
        updated_config_data = self.replace_vars_in_testcase(
            data["config_data"], self.user_global_variables)
        data["config_data"] = updated_config_data
        return data

    def getDependency(self, testcases: Dict):
        """
        yields the testcases with least dependncy first
        Reverse toplogical sort
        accept all testcases dictionary as arguments
        """
        adj_list = {}
        for key, value in testcases.items():

            dependencies=list(set(list(value.get("DEPENDENCY", "").upper().split(",")))  - set([""]))
            adj_list[key] = list(map(lambda i: i.strip(), dependencies))

        for key, value in adj_list.items():
            new_list = []
            for testcase in value:

                testcase = testcase.split(":")
                if len(testcase) > 1:
                    new_list.append(testcase[1])
                else:
                    new_list.append(testcase[0])

            adj_list[key] = set(new_list)

        while adj_list:
            """
            return testcases that doesn't depend on other testcase
            """
            top_dep = set(
                i for dependents in list(adj_list.values()) for i in dependents
            ) - set(adj_list.keys())
            top_dep.update(key for key, value in adj_list.items() if not value)

            if not top_dep:
                logging.critical(
                    "circular dependency found please remove the cirular dependency"
                )
                logging.debug("possible testcase with circular dependencies")
                sys.exit(1)

            adj_list = {key: value - top_dep for key,
                        value in adj_list.items() if value}
            result = []
            for key in testcases:
                if key in top_dep:
                    result.append(testcases[key])
            yield result

    def isDependencyPassed(self, testcase: Dict) -> bool:
        """
        cheks if the dependency is passed for the testcase or not
        """

        # split on ','
        listOfTestcases = []
        # post 1.0.4
        listOfTestcases = list(set(list(testcase.get("DEPENDENCY", "").upper().split(
            ","))) - set([""]))  # if testcase.get("DEPENDENCY", None) else listOfTestcases
        for dep in listOfTestcases:

            dep_split = list(map(lambda i: i.strip(), list(dep.split(":"))))

            if len(dep_split) == 1:
                # NAME to name, to_list()
                if dep_split[0] not in self.DATA.testcase_details["name"].to_list():
                    return 'err'
                if ((self.DATA.testcase_details[self.DATA.testcase_details["name"] == dep_split[0]]['status'].iloc[0]) != status.PASS.name):
                        return 'fail'

            else:
                if dep_split[0].upper() == "P":
                    if dep_split[1] not in self.DATA.testcase_details["name"].to_list():
                        return 'err'
                    # way to parsing the df
                    if ((self.DATA.testcase_details[self.DATA.testcase_details["name"] == dep_split[1]]['status'].iloc[0]) != status.PASS.name):
                        return 'fail'

                elif dep_split[0].upper() == "F":
                    if dep_split[1] not in self.DATA.testcase_details["name"].to_list():
                        return 'err'
                    if (
                        (self.DATA.testcase_details[self.DATA.testcase_details["name"]
                        == dep_split[1]]['status'].iloc[0])
                        != status.FAIL.name
                    ):
                        return 'fail'

                else:
                    logging.error("Wrong Dependency flag. Should be either P/F but {} was found in testcase {}".format(dep_split[0].upper(), testcase.get("NAME")))
                    return 'err'

        return 'true'

    # Function for sending total, pass, fail in order
    def totalOrder(self, unsorted_dict):

        prio_list = ['TOTAL', 'PASS', 'FAIL']
        sorted_dict = {}
        for key in prio_list:
            if key in unsorted_dict:
                sorted_dict[key] = unsorted_dict[key]
                unsorted_dict.pop(key)
            elif key == "PASS" or key == 'FAIL':
                sorted_dict[key] = 0
        sorted_dict.update(unsorted_dict)
        return sorted_dict

    def sortTestcase(self, all_config_data: Type[Dict]):
        priority_list = list()
        unpriority_list = list()
        global_vars = list()
        sorted_testcase_dict = dict()
        suite_data = all_config_data["SUITE_DATA"]
        testcase_data = all_config_data["TESTCASE_DATA"]
        for testcase, each_testcase_data in testcase_data.items():
            if (each_testcase_data.get("GLOBAL_VARS", None) is not None):
                priority_list.append(testcase)
                vars = each_testcase_data["GLOBAL_VARS"]
                vars_list = vars.split(";")
                for v in vars_list:
                    global_vars.append(v)
            else:
                unpriority_list.append(testcase)
        global_variables_dict = self.fetchGlobalVariableDetails(global_vars)
        updated_testcase_data = self.update_testcase_values(
            testcase_data, global_variables_dict)
        for testcase in priority_list:
            if ("GLOBAL_VARS" in updated_testcase_data.get(testcase).keys()):
                sorted_testcase_dict[testcase] = updated_testcase_data[testcase]
            else:
                unpriority_list.append(testcase)
        for testcase in unpriority_list:
            sorted_testcase_dict[testcase] = updated_testcase_data[testcase]
        suite_data["GLOBAL_VARIABLES"] = global_variables_dict
        all_config_data["TESTCASE_DATA"] = sorted_testcase_dict
        return all_config_data

    def fetchGlobalVariableDetails(self, global_vars: type[list]):
        self.update_global_value = list()
        filtered_global_var = dict()
        for var in global_vars:
            if "SET" in var.upper() and '$[#' in var:
                variable_name = (var.split("=")[0]).replace(".", "_").replace(
                    "set".casefold(), "").replace("$[#", "").replace("]", "").replace(" ", "")
                value = var.split("=")[1]
                filtered_global_var[variable_name.upper()] = value
                if '$[#' in value:
                    self.update_global_value.append(variable_name)
        return filtered_global_var

    def update_testcase_values(self, testcase_data, global_variables_dict):
        # optimize testcase data, if global variable has constant data then it is global_vars tag is removed from testcase
        for _, testcase_val in testcase_data.items():
            condition = list()
            flag = False
            for param, value in testcase_val.items():
                if param.casefold() == "GLOBAL_VARS".casefold():
                    flag = True
                    global_var_list = value.split(";")
                    for global_vars in global_var_list:
                        if "$[#" in global_vars.split("=")[1].strip(" "):
                            condition.append(global_vars)

                if "$[#GLOBAL." in value.replace(" ", "") and "SET$[#GLOBAL.".casefold() not in value.replace(" ", ""):
                    # global_var = value.replace("$[#", "").replace(" ", "").replace(".", "_").replace("]", "")
                    var_name = (value[value.find("$"): value.find("]")]).replace(
                        "$[#", "").replace("]", "").replace(".", "_")
                    var_value = global_variables_dict.get(var_name.upper())
                    if var_value is None or "$[#" in var_value:
                        continue
                    final_param_value = value[:value.find(
                        "$")] + var_value + value[value.find("]") + 1:]
                    testcase_val[param] = final_param_value
            if flag == True and len(condition) == 0:
                del testcase_val["GLOBAL_VARS"]
                flag = False
            elif flag == True:
                condition_str = ";".join(condition)
                testcase_val["GLOBAL_VARS"] = condition_str

        return testcase_data

    def replace_vars_in_testcase(self, testcase_config_data, global_variables):
        for key, val in testcase_config_data.items():
            if type(val) is str and val is not None and "SET$[#GLOBAL.".casefold() not in val.replace(" ", "").casefold() and "$[#GLOBAL.".casefold() in val.strip(" ").casefold():
                if val.find("$") != -1 and val.find("[") != -1:
                    temp_var = val[val.find("$"): val.find("]")]
                    global_var_name = temp_var.replace(" ", "").replace(
                        "$[#", "").replace("]", "").replace(".", "_")
                    # and "$[#" not in global_variables.get(global_var_name.upper())
                    if global_variables.get(global_var_name.upper(), None) is not None:
                        new_val_to_key = val[:val.find(
                            "$")] + str(global_variables.get(global_var_name.upper())) + val[val.find("]") + 1:]
                        testcase_config_data[key] = new_val_to_key
        return testcase_config_data            
                        



from gempyp.config.xmlConfig import XmlConfig
import argparse
from gempyp.engine.engine import Engine
from gempyp.libs.common import download_common_file
from gempyp.data_uploader import dataUploader
import logging

class Gempyp:
    def __init__(self):
        """
        initiating some variables  
        """
        self.config = None
        self.MAIL = None
        self.PROJECT_NAME = None
        self.REPORT_NAME = None
        self.MODE = None
        self.ENVIRONMENT = None
        self.args = {}
        self.THREADS = None
        self.JEWEL_BRIDGE_TOKEN = None 
        self.REPORT_LOCATION = None
        self.CATEGORY=None
        self.JEWEL_USER=None
        self.S_RUN_ID = None
        self.TESTCASE_LIST = None
        self.ENTER_POINT = None
        self.S_ID = None
        self.RUN_TYPE = None
        self.RUN_MODE = None
        self.RE_RUN = None
    
    def argParser(self):
        """Argument parser to help running through CLI"""

        parser = argparse.ArgumentParser()
        parser.add_argument('-config','-c',dest='config',type=str, required=False, help='Path of config file. Accepts both absolute and relative path.')
        parser.add_argument('-mailTo','-mailTo',dest='MAIL-TO', type=str, required=False, help='Specify recipient email addresses.Can send multiple email id by comma separated.')
        parser.add_argument('-mailCc','-mailCc',dest='MAIL-CC', type=str, required=False, help='Provide email addresses for carbon copy (cc).Can send multiple email id by comma separated.')
        parser.add_argument('-mailBcc','-mailBcc',dest='MAIL-BCC', type=str, required=False, help='Specify BCC email addresses.Can send multiple email id by comma separated.')
        parser.add_argument('-project','-p',dest='PROJECT_NAME', type=str, required=False, help='Specify project name.')
        parser.add_argument('-Report_name','-rn',dest='REPORT_NAME', type=str, required=False, help='Specify report name.')
        parser.add_argument('-mode','-mode',dest='MODE', type=str, required=False, help='Specify execution mode. Values can be sequence and optimize.')
        parser.add_argument('-env','-env',dest='ENVIRONMENT', type=str, required=False,help='Specify the testing environment.' )
        parser.add_argument('-threads','-t',dest='THREADS',type=int, required=False, help='Number of threads use in optimoize mode. This should be a numeric value.')
        parser.add_argument('-bridge_token','-token',dest='JEWEL_BRIDGE_TOKEN',type=str, required=False, help='Jewel bridge token.')
        parser.add_argument('-username','-username',dest='JEWEL_USER',type=str, required=False, help='Jewel username.')
        parser.add_argument('-report_location','-rl',dest='REPORT_LOCATION',type=str, required=False, help='Location for jewel report. Can ba absolute and relative path.')
        parser.add_argument('-category','-category',dest='CATEGORY',type=str, required=False, help='Category of testcases you want to run.')
        parser.add_argument('-run_id','-run_id',dest='S_RUN_ID',type=str, required=False, help='For specifying s_run_id.')
        parser.add_argument('-tc','-testcase_list',dest='TESTCASE_LIST',type=str, required=False, help='Name of testcases user want to run. User can pass multiple testcases in coma separated form.')
        parser.add_argument('-enter_point','-enter_point',dest='ENTER_POINT',type=str, required=False, help='Base Url user want to use for API. This should be a url.')
        parser.add_argument('-s_id','-s_id',dest='S_ID',type=str, required=False)
        parser.add_argument('-run_type','-run_type',dest='RUN_TYPE',type=str, required=False)
        parser.add_argument('-run_mode','-run_mode',dest='RUN_MODE',type=str, required=False)
        parser.add_argument('-rerun','-rerun',dest='RE_RUN',type=str, required=False, help='Specify the path of json file you want to create report for. This should json file and created by gempyp in case of Apis failure.')

        args = parser.parse_args()
        return args

    def runner(self):
        """
        This function takes the config and updates the config data in case or cli run and direct(python) run
        """
        s_run_id = vars(self)["S_RUN_ID"]
        if self.args.get('RE_RUN',None) != None:
                print("Trying to Reupload Data")
                dataUploader(self.args.get('RE_RUN'),self.args.get('JEWEL_BRIDGE_TOKEN',None))
        elif self.RE_RUN != None:
                print("Trying to Reupload Data")
                dataUploader(self.RE_RUN, self.JEWEL_BRIDGE_TOKEN)
        # if self.args.RE_RUN
        else:
            file_path=download_common_file(self.config)
            config=XmlConfig(file_path,s_run_id)
            if not self.args:
                # del self.__dict__["args"]
                config.cli_config = vars(self)
            else:
                config.cli_config = self.args
            config.update()
            Engine(config)

    def parser(self):
        """Calls the parser and handles the case of no cli args"""
        args = self.argParser()
        if args.config != None:
            self.config = args.config
        self.args = vars(args)
        self.runner()

def main():
    obj = Gempyp()
    #obj.config = "C:\\Users\\an.pandey\\gempyp\\tests\\configTest\\Gempyp_Test_Suite.xml"
    #obj.ENV = ""
    #obj.MAIL = ""
    obj.parser()

if __name__ == "__main__":
    obj = Gempyp()
    
    #obj.config = "C:\\Users\\an.pandey\\gempyp\\tests\\configTest\\Gempyp_Test_Suite.xml"
    #obj.ENV = ""
    #obj.MAIL = ""
    obj.parser()
    




from typing import Dict, List
from gempyp.libs.common import remove_none_from_dict


def xmlToDict(root) -> Dict:
    """convert data from xml to dict"""
    Result = {}
    for element in root:
        # if element has children
        if len(element) != 0:
            # check if the element childern are list or dict

            # if element has only 1 children
            # or if first 2 elements are not same we treat them as dict
            if len(element) == 1 or element[0].tag.upper() != element[1].tag.upper():
                Result[element.tag.upper()] = xmlToDict(element)

            # we treat them as list
            else:
                Result[element.tag.upper()] = xmlToList(element)

        # if the element doesnot have any children
        else:
            if(type(element.tag)==str):
                Result[element.tag.upper()] = element.text
    return remove_none_from_dict(Result)  ##################### post 1.0.4


def xmlToList(root) -> List:
    Result = []

    for element in root:
        # if element has children
        # same logic as xmlToDict
        if len(element) != 0:
            if len(element) == 1 or element[0].tag != element[1].tag:
                Result.append(xmlToDict(element))

            # we treat them as list
            else:
                Result.append(xmlToList(element))

        else:
            Result.append(element.text)

    return Result



import pandas as pd
from gempyp.libs.common import get_reason_of_failure
import traceback
from gempyp.libs.enums.status import status


# for handling data formatting
def dateFormatHandling(src_df,tgt_df):
    try:
        src_df = formatingDate(src_df)
        tgt_df = formatingDate(tgt_df) 
        # print("tgt_df",tgt_df.dtypes)
        return src_df, tgt_df
    except Exception:
        raise Exception

def formatingDate(df):
    columns = df.select_dtypes(include=['datetime64']).columns
    for column in list(columns):
        df[column] = pd.to_datetime(df[column],errors='ignore')
    return df

def columnCompare(src_df, tgt_df, keys, compare_keys):
    """we are trying to get list of columns which we dont want to compare and droping them"""
    columns = keys+compare_keys
    src_df = src_df[columns]
    tgt_df = tgt_df[columns]
    return src_df, tgt_df

def skipColumn(skip_column,src_df,tgt_df,key,reporter):
            try:
                skip_column = skip_column.split(',')
                src_columns = src_df.columns
                tgt_columns = tgt_df.columns
                src_skip_columns = list(set(src_columns) & set(skip_column))
                tgt_skip_columns = list(set(tgt_columns) & set(skip_column))
                flag = False
                for i in key:
                    if i in skip_column:
                        flag = True
                        break
                if flag:
                    reporter.addRow("Columns given for skip are present in keys tag","Aborting skip columns",status.INFO)
                    return src_df,tgt_df
                else:
                    # for i in skip_column:
                    src_df.drop(src_skip_columns,axis=1,inplace=True)
                        # del src_df[i]
                    # for i in skip_column:
                    tgt_df.drop(tgt_skip_columns,axis=1,inplace=True)
                        # del tgt_df[i]
                    # headers = list(set(headers)-set(skip_column))
                    return src_df,tgt_df
            except Exception as e:
                reporter.addRow("While skipping columns",get_reason_of_failure(traceback.format_exc(), e),status.FAIL)
                return src_df, tgt_df



from gempyp.pyprest.keyCheck import KeyCheck
from gempyp.pyprest import utils
from copy import deepcopy
import logging as logger
from gempyp.pyprest.variableReplacement import VariableReplacement as var_replacement
from gempyp.pyprest.preVariables import PreVariables



class PostVariables:
    def __init__(self, pyprest_obj):
        self.pyprest_obj = pyprest_obj
        self.logger = self.pyprest_obj.logger
        self.isLegacyPresent = hasattr(self.pyprest_obj, "legacy_res")
        # get variable written in data["POST_VARIABLE"]
        # postdefined func
        # remove $[#]
        # assign_values and append to a dict
        # different dicts for loal and suite variables
        

    def postVariables(self):
        self.logger.info("************** INSIDE POST VARIABLES  **************")
        post_variables_str = self.pyprest_obj.post_variables
        global_variables = self.pyprest_obj.data.get("GLOBAL_VARIABLES")
        testcase_global_vars = self.pyprest_obj.data.get("config_data")
        if testcase_global_vars.get("GLOBAL_VARS", None) is not None:
            post_variables_str += ";" + testcase_global_vars.get("GLOBAL_VARS")
        if post_variables_str:

            # separate by ;
            variable_list = post_variables_str.split(";")
            for each in variable_list:
                each_item = each.split("=")
                scope = "local"
                # checking for syntax
                if "SET" in each_item[0].upper() and '$[#' in each_item[0]:
                    # key = each_item[0].strip("set $[#").strip("Set $[#").strip("SET $[#").strip("]")
                    key = each_item[0].split("#")[1].replace("]","")

                    # find suite variables
                    if "SUITE." in key.upper():
                        scope = "suite"
                        key = key.replace(".", "_")     
                    if "SUITE." in str(each_item[0].strip(" ")):
                        key = "SUITE_" + each_item[0].strip(" ").replace("set $[#SUITE.","").replace("]","").upper()
                        
                        self.pyprest_obj.variables["suite"][key] = PreVariables(self.pyprest_obj).getFunctionValues(each_item[1].strip())
                    if "GLOBAL." in str(each_item[0].strip(" ")):
                        scope = "GLOBAL_VARIABLES"
                        key = "GLOBAL_" + each_item[0].strip(" ").replace("set $[#GLOBAL.","").replace("]","").upper()
                        val = PreVariables(self.pyprest_obj).getFunctionValues(each_item[1].strip())
                    # check for postdefined functions and response variables
                    if "$[#" in each_item[1].strip(" "):    
                        # check for predefined function
                        self.pyprest_obj.variables[scope][key] = PreVariables(self.pyprest_obj).getFunctionValues(each_item[1].strip())
                        
                        # if not found in predefined functions, check in response
                        response_key = each_item[1].strip().strip("$[#").strip("]")
                        # find key in response  
                        # call keycheck
                        response_key_partition = response_key.split(".")
                        response_json = utils.formatRespBody(self.pyprest_obj.res_obj.response_body)
                        if self.isLegacyPresent and 'legacy' in response_key_partition:
                            response_json = utils.formatRespBody(self.pyprest_obj.legacy_res.response_body)    
                        result = KeyCheck(self.pyprest_obj).findKeys(response_json, deepcopy(response_key_partition), deepcopy(response_key_partition))
                        # if result is not "FOUND" then can't set value
                        if result.upper() != "FOUND":
                            self.logger.info("====== Key Not Found in response =======")
                            self.logger.info("'" + each_item[1].strip() + "' is not found")

                            # check predefined functions
                            self.pyprest_obj.variables[scope][key] = PreVariables(self.pyprest_obj).getFunctionValues(each_item[1].strip())
                        else:
                            new_list = utils.fetchValueOfKey(response_json, response_key_partition, result)
                            self.pyprest_obj.variables[scope][key] = new_list[response_key]
                            self.logger.info(self.pyprest_obj.variables[scope][key])
                            if (scope.casefold() == "GLOBAL_VARIABLES".casefold()):
                                self.logger.info("Updating the global variable -- {var}".format(var=key))
                                self.pyprest_obj.__dict__["data"]["GLOBAL_VARIABLES"][key] = new_list[response_key]
                                global_variables[key] = new_list[response_key]
                                self.logger.info(self.pyprest_obj.__dict__["data"]["GLOBAL_VARIABLES"][key])
                                if self.pyprest_obj.__dict__["data"]["GLOBAL_VARIABLES"].get("UPDATED_GLOBAL_VARS", None) is None:
                                    self.pyprest_obj.__dict__["data"]["GLOBAL_VARIABLES"]["UPDATED_GLOBAL_VARS"] = list()
                                    self.pyprest_obj.__dict__["data"]["GLOBAL_VARIABLES"]["UPDATED_GLOBAL_VARS"].append(key)
                                else:
                                    self.pyprest_obj.__dict__["data"]["GLOBAL_VARIABLES"]["UPDATED_GLOBAL_VARS"].append(key)

                    # key not found in response, checking pre variables and pre variables
                    if "$[#" in each_item[1].strip(" "):
                        var_replacement(self.pyprest_obj).variableReplacement()
            self.logger.info(f"variables dict after setting POST VARIABLES: -------- {str(self.pyprest_obj.variables)} ")


import os
from gempyp.engine import dataUpload
import traceback
import logging
import re
import sys

# have to discuss about the default location
DEFAULT_GEMPYP_FOLDER = os.getcwd()
backup_data=None
DEBUG = True
THREADS = 8
encrypt_key = b'sEKTykqMLP_iCwlMtiBR_9SQ0v9N1OT3ajVAAaI4AkQ='
apiSuccess = False
project_id = "Test_id"
default_baseurl="https://apis.gemecosystem.com/gemEcosystemDashboard/v1"
urls = {"data":{} }
# for getting urls using url tag from config file
def getEnterPoint(url, bridge_token, user_name):
    global urls
    try:
        url = checkUrl(url)
        response = dataUpload._sendData(" ", url, bridge_token, user_name,"GET")
        if response.status_code == 200:
            url_enter_point = response.json()
            urls["data"]=url_enter_point["data"]
            global apiSuccess
            apiSuccess = True
        elif re.search('50[0-9]',str(response.status_code)):
            logging.warning("Error Occurs While Getting the BASE_URLs")
            logging.warning("Response status code: {}".format(response.status_code))
            logging.warning("Response: {}".format(response.text))
        else:
            logging.info("Some Error From the Client Side, Maybe username or bridgeToken, Therefore terminating execution")
            logging.warning("Response status code: {}".format(response.status_code))
            logging.warning("Response: {}".format(response.text))
            # sys.exit()
    except Exception as e:
            traceback.print_exc()
            logging.warning("Error Occurs While Getting the BASE_URLs")
            logging.warning("Response status code: {}".format(response.status_code))
            logging.warning("Response: {}".format(response.text))
            sys.exit()

# for sending urls to dataupload file
def getUrls(apiName):
    return urls["data"].get(apiName, None)

def checkUrl(url):
    try:
        # l = url.index('.com')
        # url = url[:l+4:] + "/enter-point"
        if(url.__contains__("/enter-point")):
            return url
        else:
            return url.strip("/")+"/enter-point"
    except Exception:
        traceback.print_exc()
        logging.warning("Error Occurs While handling the BASE_URLs")

count = 0




from gempyp.pyprest.restObj import RestObj


# create an object of PYPREST helper and take that object in your before and after methods,then return it.

class before:
    def before_(self, obj):
        print("---------------------------------inside before method--------------------------------------")
        print("PROJECT", obj.project)
        obj.project = "ABC"
        return obj


if __name__ == "__main__":
    obj = before().before_(RestObj(project="SOME_PROJECT"))


from gempyp.engine.baseTemplate import TestcaseReporter
import glob, configparser, os, uuid, platform, tempfile, json, logging
from gempyp.libs.enums.status import status
from gempyp.engine.testData import TestData
from gempyp.engine.runner import getOutput
from gempyp.libs.logConfig import my_custom_logger, LoggingConfig
from gempyp.engine import dataUpload
from gempyp.libs.common import *
from datetime import datetime, timezone
from gempyp.libs.enums.status import status
import time

class Common:
    DATA = TestData()
    data = {}
    instance = None
    _dict = {}

    @classmethod
    def get_instance(self):
        testcase_name = "UI_Automation"
        # testcase_name = re.sub('[^A-Za-z0-9]+', '_', testcase_name)
        log_path = "ui_automation.txt"
        # if len(log_path) > 240:
        #     log_path = os.path.join(os.environ.get('TESTCASE_LOG_FOLDER'), str(uuid.uuid4()) +'.txt')
        self.logging = my_custom_logger(log_path)
        LoggingConfig(log_path)
        if self.instance is None:
            self.instance = Common()
        return self.instance


    def makeSuiteDetails(self, plugin_data:dict)->None:
            """
            this function will help to create suitedetails that will be used in sendSuiteData API
            plugin_data will be dictionary that holds the data from the hooks (in behave we are using it mainly for expectedTestcase count)
            this function is mainly picking data from .ini file(currently pytest.ini from root directory)
            """          

            try:  
                path = glob.glob(os.path.join('**','pytest.ini'),recursive=True)
                config = configparser.ConfigParser()
                config.read(path)
                self.dataFromIni = dict(config.items('pytest'))
                self.s_run_id = self.dataFromIni.get("project-name") + "_" + self.dataFromIni.get("environment","beta") + "_" + str(uuid.uuid4())
                self.dataFromIni['s-run-id'] = self.s_run_id
                run_mode = "LINUX_CLI"
                if os.name == 'nt':
                    run_mode = "WINDOWS"
                Suite_details = {
                    "s_run_id": self.dataFromIni["s-run-id"],
                    "s_start_time": datetime.now(timezone.utc),
                    "s_end_time": None,
                    "s_id": self.dataFromIni.get("S_ID", "test_id"),
                    "status": status.EXE.name,
                    "project_name": self.dataFromIni.get("project-name",""),
                    "report_name": self.dataFromIni.get("report-name","smoke_test"),
                    "run_type": "ON DEMAND",
                    "user": self.dataFromIni.get('jewel-user',None),
                    "env": self.dataFromIni.get("environment","beta"),
                    "machine": self.dataFromIni.get("machine",platform.node()),
                    "run_mode": run_mode,
                    "os": platform.system().upper(),
                    "meta_data": [],
                    "testcase_info": None,
                    "expected_testcases":plugin_data.get('expected-testcases',1),
                    "framework_name": "GEMPYP",
                }
                self.DATA.suite_detail = self.DATA.suite_detail.append(
                    Suite_details, ignore_index=True
                )
            except Exception:
                traceback.print_exc()


    def getTestcaseData(self, testcaseData:dict)->None:
            """
            this function is mainly to create testcase data 
            in this function we sending testcaseName key
            """
            try:
                self.jewel_user=False
                # global reporter
                self.reporter = TestcaseReporter(self.dataFromIni.get("project-name",""), testcaseData.get("testcaseName","Demo Test"))
                # reporter = TestcaseReporter(self.dataFromIni.get("project-name",""), testcaseData.get("testcaseName","Demo Test"))
                self.reporter.Status = status
                # reporter.Status = status
                projectName = self.dataFromIni.get("project-name","")
                env = self.dataFromIni.get("environment","beta")
                testcaseName = testcaseData.get("testcaseName","Demo Test")
                self.data["JEWEL_USER"] = self.dataFromIni.get('jewel-user',None)
                self.data["JEWEL_BRIDGE_TOKEN"] = self.dataFromIni.get("jewel-bridge-token")
                self.data["REPORT_LOCATION"] = self.dataFromIni.get("report-location")
                self.data["MACHINE"] = self.dataFromIni.get("machine",platform.node())
                self.data["MAIL_TO"] = self.dataFromIni.get("mail-to")
                self.data["MAIL_CC"] = self.dataFromIni.get("mail-cc")
                self.data["MAIL_BCC"] = self.dataFromIni.get("mail-bcc")
                self.data["ENTER_POINT"] = self.dataFromIni.get("enter-point")
                if self.data["JEWEL_USER"] and self.data["JEWEL_BRIDGE_TOKEN"]:
                    self.jewel_user=True
                report_type = self.data["REPORT_NAME"] = self.dataFromIni.get("report-name","Smoke Test")
                if not self.dataFromIni.get("s-run-id"):
                    self.data["s_run_id"] = self.s_run_id
                    os.environ["s_run_id"] = self.s_run_id
                else:
                    self.s_run_id = self.data["s_run_id"] = self.dataFromIni.get("s-run-id")
            except Exception:
                traceback.print_exc()

    def sendSuiteData(self):
        try:
            """this function is for calling first post api for suite data"""
            runBaseUrls(True,self.dataFromIni.get("enter-point"),self.dataFromIni.get("jewel-user"),self.dataFromIni.get("jewel-bridge-token"))
            dataUpload.sendSuiteData((self.DATA.toSuiteJson()), self.dataFromIni.get("jewel-bridge-token"), self.dataFromIni.get("jewel-user")) # check with deamon, should insert only once
        except Exception:
                traceback.print_exc()


    def send_testcase_data(self):
        try:
            # test_func_node = item.parent.obj
            output = []
            # global reporter
            self.reporter.finalizeReport()
            # reporter.finalizeReport()
            # create testcase self.reporter json
            # global reporter
            self.reporter.json_data = self.reporter.template_data.makeTestcaseReport()
            # reporter.json_data = reporter.template_data.makeTestcaseReport()
            # serializing data, adding suite data
            report_dict = self.reporter.serialize()
            # report_dict = reporter.serialize()
            report_dict["TESTCASEMETADATA"] = self.getMetaData()
            report_dict["config_data"] = self.getConfigData()
            # creating output json
            output.append(getOutput(report_dict))
            output[0]['testcase_dict']['run_type'], output[0]['testcase_dict']['run_mode'],output[0]['testcase_dict']['job_name'],output[0]['testcase_dict']['job_runid']="ON DEMAND","WINDOWS",None,None
            for i in output:
                i["testcase_dict"]["steps"] = i["json_data"]["steps"]
                dict_ = {}
                dict_["testcases"] = {}
                dict_["REPORT_LOCATION"] = os.getenv("REPORT_LOCATION")
                dict_["misc_data"] = {}
                # self.tmp_dir = os.path.join(tempfile.gettempdir(), self.data["s_run_id"] + ".txt")
                self.DATA.testcase_details = self.DATA.testcase_details.append(
                    i["testcase_dict"], ignore_index=True
                )   
                # self.self.DATA.testcaseDetails = pd.concat([self.self.DATA.testcaseDetails, pd.DataFrame(list(i["testcase_dict"].items()))])
                self.updateTestcaseMiscData(i["misc"], tc_run_id=i["testcase_dict"].get("tc_run_id"))
                suite_data = self.DATA.getJSONData()
                if isinstance(suite_data, str):
                    suite_data = json.loads(suite_data)
                if len(self._dict)==0:
                    dict_[self.s_run_id] = self.updateSuiteData(suite_data)
                    dict_["testcases"][i["testcase_dict"].get("tc_run_id")] = i["json_data"]
                    self._dict = dict_
                else:
                    self._dict["s_run_id"] = self.updateSuiteData(suite_data,self._dict[self.s_run_id])
                    self._dict["testcases"][i["testcase_dict"].get("tc_run_id")] = i["json_data"]
                updatedData=json.loads(self.DATA.totestcaseJson(i["testcase_dict"]["tc_run_id"].upper(), self.data["s_run_id"]))
                logging.info("TC_RUN_ID : "+i["testcase_dict"]["tc_run_id"].upper())
                dataUpload.sendTestcaseData(json.dumps(updatedData), self.data["JEWEL_BRIDGE_TOKEN"], self.data["JEWEL_USER"])  # instead of output, I need to pass s_run id and  tc_run_id
                
        except Exception:
            traceback.print_exc()

    def getMetaData(self):
        try:
            data = {
                'PROJECT_NAME': self.dataFromIni.get("project-name",""), 
                'ENVIRONMENT': self.dataFromIni.get("environment",""), 
                's_run_id': self.dataFromIni.get("s-run-id",""), 
                'USER': self.dataFromIni.get("jewel-user",""), 
                'MACHINE': self.dataFromIni.get("machine",platform.node()), 
                'REPORT_LOCATION': self.dataFromIni.get("report-location",""),
                'SUITE_VARS': self.dataFromIni.get("suite-vars", {}),
                'INVOKE_USER': os.getenv("INVOKEUSER", self.data["JEWEL_USER"])}
            return data
        except Exception:
            traceback.print_exc()


    def getConfigData(self):
        try:
            data = {'NAME': self.dataFromIni["project-name"], 'CATEGORY': 'External','LOGGER': ""}
            return data
        except Exception:
            traceback.print_exc()

    def updateTestcaseMiscData(self, misc, tc_run_id):
            
        try:
            """
            updates the misc data for the testcases
            """
            miscList = []

            for misc_data in misc:
                temp = {}
                # storing all the key in upper so that no duplicate data is stored
                temp["key"] = misc_data.upper()
                temp["value"] = misc[misc_data]
                temp["run_id"] = tc_run_id
                temp["table_type"] = "TESTCASE"
                miscList.append(temp)

            self.DATA.misc_details = self.DATA.misc_details.append(
                miscList, ignore_index=True
            )
        except Exception:
            traceback.print_exc()


    def updateSuiteData(self, current_data, old_data=None):
        try:
            """
            updates the suiteData after all the runs have been executed
            """
            
            start_time = current_data["suits_details"]["s_start_time"]
            end_time = current_data["suits_details"]['testcase_details'][-1]["end_time"]
            testcaseData = current_data["suits_details"]['testcase_details']
            if old_data:
                testcaseData = (old_data["suits_details"]['testcase_details']
                + current_data["suits_details"]['testcase_details'])
                start_time = old_data["suits_details"]["s_start_time"]
            statusDict = {k.name: 0 for k in status}
            for i in testcaseData:
                statusDict[i["status"]] += 1
            # get the status count of the status
            SuiteStatus = status.FAIL.name

            # based on the status priority
            for s in status:
                if statusDict.get(s.name, 0) > 0:
                    SuiteStatus = s.name

            current_data["suits_details"]["status"] = SuiteStatus
            current_data["suits_details"]['testcase_details'] = testcaseData
            current_data["suits_details"]["s_start_time"] = start_time
            current_data["suits_details"]["s_end_time"] = end_time
            count = 0
            for key in list(statusDict.keys()):
                if statusDict[key] == 0:
                    del statusDict[key]
                else:
                    count += statusDict[key]
            statusDict["total"] = count
            current_data["suits_details"]["expected_testcases"]=count
            current_data["suits_details"]["testcase_info"] = statusDict

            return current_data
        except Exception:
            traceback.print_exc()
    
    def createReport(self):
        
        try:    
            json_data = self._dict[self.s_run_id]
            jewel= self.jewel_user
            bridgetoken=self.dataFromIni.get("jewel-bridge-token")
            testcaseData = self._dict["testcases"]
            suite_data = json_data["suits_details"]
            suite_data["misc_data"] = self._dict["misc_data"]
            username = suite_data["user"]
            del suite_data["testcase_details"]
            del suite_data["testcase_info"]
            response_code = dataUpload.sendSuiteData(json.dumps(suite_data), bridgetoken, username, mode="PUT")
            if response_code in [200,201]:
                jewelLink = DefaultSettings.getUrls('jewel-url')
                if jewelLink is not None:
                    # jewel_link = f'{jewelLink}/#/autolytics/execution-report?s_run_id={self.s_run_id}&p_id={DefaultSettings.project_id}'
                    jewel_link = f'{jewelLink}/#/autolytics/execution-report?s_run_id={self.s_run_id}'
                    logging.info(f"Jewel link of gempyp report - {jewel_link}")
            else:
                logging.info("Some error occurred while uploading data")
        except Exception:
            traceback.print_exc()
    

common = Common.get_instance()


import os
import logging
from typing import Dict
from datetime import datetime, timezone
from itertools import chain
import json
from gempyp.libs.enums.status import status
from gempyp.libs.common import findDuration, dateTimeEncoder
import traceback
from gempyp.libs.gem_s3_common import upload_to_s3, create_s3_link
from gempyp.config import DefaultSettings

class TemplateData:
    def __init__(self, header="Gemini Report"):
        # initliza the data to be stored as a JSON
        self.REPORTDATA = {"Header": header, "steps": []}

    def newReport(self, project_name: str, tescase_name: str):
        metadata = []
        self.kwargs_list=[]
        # 1st Column
        column1 = {
            "TESTCASE NAME": tescase_name,
            "SERVICE PROJECT": project_name,
            "DATE OF EXECUTION": {"value": datetime.now(timezone.utc), "type": "date"},
        }
        metadata.append(column1)

        self.REPORTDATA["meta_data"] = metadata

    def newRow(self, title: str, description: str, status: status, **kwargs):
        step = {"Step Name": title, "Step Description": description, "status": status}
        if not kwargs.get("attachment"):
            kwargs.pop("attachment")
        if kwargs:
            if kwargs not in self.kwargs_list:
                self.kwargs_list.append(kwargs)
        step.update(kwargs)
        self.REPORTDATA["steps"].append(step)

    # finalize the result. Calculates duration etc.
    def finalizeResult(
        self, begin_time: datetime, end_time: datetime, status_counts: Dict
    ):
        # column2
        column2 = {
            "EXECUTION STARTED ON": {"value": begin_time, "type": "datetime"},
            "EXECUTION ENDED ON": {"value": end_time, "type": "datetime"},
            "EXECUTION DURATION": findDuration(begin_time, end_time),
        }

        # column3
        column3 = {k.name: v for k, v in status_counts.items()}
        self.REPORTDATA["meta_data"].append(column2)
        self.REPORTDATA["meta_data"].append(column3)
        # filters
        # self.REPORTDATA["FilterNames"] = self._getFilters()
        # filter_values = {}
        # filter_values["status"] = [value.name for value in status_counts.keys()]
        # self.REPORTDATA["FilterValues"] = filter_values

    """def _getFilters(self) -> Dict:
        
        # return the unique columns

        filter_names = list(
            set(chain.from_iterable(step.keys() for step in self.REPORTDATA["steps"]))
        )
        # filter_names.pop("status")
        filter_dict = {name: "Input" for name in filter_names}
        filter_dict["status"] = "Dropdown"

        return filter_dict"""

    # Converts the data to the JSON
    def _toJSON(self) -> str:
        """
        dump the data in REPORTDATA
        """
        kwargs_list_comm =  dict(j for i in self.kwargs_list for j in i.items())
        lis = kwargs_list_comm.keys()
        # for i in self.REPORTDATA["steps"]:
        #     for j in lis:
        #         if j in i:
        #             continue
        #         else:
        #             i.update({j:"-"})

        try:
            result_data = json.dumps(self.REPORTDATA, cls=dateTimeEncoder)
            return result_data
        except TypeError as error:
            logging.error("Error occured while serializing the testcase Result Data")
            logging.error(f"Error: {error}")
        except Exception as e:
            logging.error("some Error occured")
            logging.error(f"Error: {e}")
        return "Error"
    
    def makeSuiteReport(self, json_data, testcase_data, ouput_folder,jewel_user):
        """
        saves the report json 
        """
        suiteReport = None
        date = datetime.now().strftime("%Y_%b_%d_%H%M%S_%f")

        suite_path = os.path.dirname(__file__)
        suite_path = os.path.join(os.path.split(suite_path)[0], "final_report.html")
        with open(suite_path, "r") as f:
            suiteReport = f.read()

        reportJson = json_data
        reportJson = json.loads(reportJson)
        reportJson["teststep_details"] = testcase_data
        repJson = reportJson
        # self.testcaseData = json.dumps(self.testcaseData)
        reportJson = json.dumps(reportJson)
        suiteReport = suiteReport.replace("DATA_1", reportJson)
        
        # if not len(testcase_data) > 0:   ##TODO
        #     return repJson, None
        ouput_file_path=""
        # if not jewel_user:
        ResultFile = os.path.join(ouput_folder, "Result_{}.html".format(date))
        ouput_file_path = ResultFile
        with open(ResultFile, "w+") as f:
            f.write(suiteReport)
        # return repJson, ouput_file_path
        return repJson, ouput_file_path
    
    def makeTestcaseReport(self):
        index_path = os.path.dirname(__file__)
        index_path = os.path.join(os.path.split(index_path)[0], "testcase.html")
        with open(index_path, "r") as f:
            Result_data = f.read()
        json_data = self._toJSON()
        return json.loads(json_data)
    
    def createLogFile(self,folder_path,output_file,suite_log_file):
        with open(output_file, 'a') as outfile:
            with open(suite_log_file, 'r') as infile:
                        # Read the contents of the file
                    file_contents = infile.read()
                    outfile.write(file_contents)
                    outfile.write('\n')
            # Iterate over all files in the folder
            for file_name in os.listdir(folder_path):
                if file_name.endswith('.txt'):
                    file_path = os.path.join(folder_path, file_name)
                    
                    # Open each file in the folder
                    with open(file_path, 'r') as infile:
                        # Read the contents of the file
                        file_contents = infile.read()
                        
                        # Append the contents to the output file
                        outfile.write(file_contents)
                        outfile.write('\n')  # Add a newline after each file (optional)
        return output_file


    def repSummary(self, repJson,jewel_link, unuploaded_path,folder_path,output_file,bridgeToken,username,suite_log, output_file_path):
        """
        logging some information
        """
        try:
            logging.info("---------- Finalised the report --------------")
            logging.info("============== Run Summary =============")
            count_info = {key.lower(): val for key, val in repJson['suits_details']['testcase_info'].items()}
            log_str = f"Total Testcases: {str(count_info.get('total', 0))} | Passed Testcases: {str(count_info.get('pass', 0))} | Failed Testcases: {str(count_info.get('fail', 0))} | "
            status_dict = {"info": "Info", "warn": "WARN", "exe": "Exe", "err":"ERR"}
            for key, val in count_info.items():
                if key in status_dict.keys():
                    log_str += f"{status_dict[key.lower()]} Testcases: {val} | "
            logging.info(log_str.strip(" | "))
            output_file=self.createLogFile(folder_path,output_file,suite_log)
            # s3_log_file_url=None
            # try:
            #     s3_log_file_url = create_s3_link(url=upload_to_s3(DefaultSettings.urls["data"]["bucket-file-upload-api"], bridge_token=bridgeToken, username=username, file=output_file,tag="public")[0]["Url"])
            # except Exception as e:
            #     logging.warn(str(e))

            if unuploaded_path != None:
                logging.info(f"Unuploaded Data File Path:{unuploaded_path}")
            #  we will check if output report is None, then display report not created
            if len(jewel_link)>0:
                # s3_report_file_url= create_s3_link(url=upload_to_s3(DefaultSettings.urls["data"]["bucket-file-upload-api"], bridge_token=bridgetoken, username=username, file=output_file_path.split('/')[-1],tag="public")[0]["Url"]) 
                logging.info('Report at Jewel: {link}'.format(link = jewel_link))
#                 logging.info('Report at S3: {link}'.format(link = s3_report_file_url))
                
            # if not jewel_user:
            logging.info('-------- Report created Successfully at: {path}'.format(path=output_file_path))


        except Exception as e:
            logging.error(traceback.format_exc(e))



import importlib
from gempyp.libs.common import readPath, download_common_file, get_reason_of_failure
from gempyp.libs.enums.status import status
from gempyp.dv.dvDatabases import Databases
import traceback
import pandas as pd
import mysql.connector
import snowflake.connector
import pg8000
import re
import time
import pymongo
import json
import ast


class Dataframe:
    def __init__(self) -> None:
        pass

    def getSourceDataFrame(self, data, logger, reporter, sourceCred):
        """checking whether data is present in csv form or in db"""
        self.configData = data.get("config_data")
        self.reporter = reporter
        self.logger = logger
        if 'SOURCE_CSV' in self.configData:
            try:
                logger.info("Getting Source_CSV File Path")
                path = self.configData['SOURCE_CSV']
                sourceCsvPath = readPath(path)
                file_obj = download_common_file(
                    sourceCsvPath, data.get("SUITE_VARS", None))
                sourceDelimiter = self.configData.get(
                    'SOURCE_DELIMITER', ',')
                source_df, source_columns = self.csvFileReader(
                    file_obj, sourceDelimiter, "source")
                return source_df, source_columns
            except Exception as e:
                reporter.addRow(
                    "Parsing source file path", "Exception occurred", status.ERR)
                reporter.addMisc(
                    "REASON OF FAILURE", get_reason_of_failure(traceback.format_exc(), e))
                raise Exception(e)
        else:
            """Connecting to sourceDB"""
            if sourceCred == None:
                # reporter.addMisc(
                    # "REASON OF FAILURE", "Source_DB or Source Connection tag not found")
                reporter.addRow("Getting source connection details",
                                "Recheck source_DB tag and source connection tag", status.ERR)
                raise Exception(
                    "Source_db tag or source connection tag not found")
            else:
                source_df, source_columns = self.dbOperations(
                    sourceCred, "SOURCE")
                return source_df, source_columns

    def getTargetDataframe(self, data, logger, reporter, targetCred):
        self.configData = data.get("config_data")
        self.reporter = reporter
        self.logger = logger
        if 'TARGET_CSV' in self.configData:
            try:
                logger.info("Getting Target_CSV File Path")
                path = self.configData['TARGET_CSV']
                targetCsvPath = readPath(path)
                file_obj = download_common_file(
                    targetCsvPath, data.get("SUITE_VARS", None))
                targetDelimiter = self.configData.get(
                    'TARGET_DELIMITER', ',')
                target_df, target_columns = self.csvFileReader(
                    file_obj, targetDelimiter, "Target")
                return target_df, target_columns
            except Exception as e:
                reporter.addRow(
                    "Parsing target file path", "Exception occurred", status.FAIL)
                # reporter.addMisc(
                #     "REASON OF FAILURE", get_reason_of_failure(traceback.format_exc(), e))
                raise Exception(e)
        else:
            """Connecting to TargetDB"""
            if targetCred == None:
                # reporter.addMisc(
                #     "REASON OF FAILURE", "Target_DB or Target Connection tag not found")
                reporter.addRow("Getting target connection details",
                                "Recheck for target_db or target connection tag", status.ERR)
                raise Exception(
                    "Target_db or target connection tag not found")
            else:
                target_df, target_columns = self.dbOperations(
                    targetCred, "TARGET")
                return target_df, target_columns

    def setConnection(self, lib, conn):
        lib = importlib.import_module(lib)
        conn = getattr(lib, conn)

    def csvFileReader(self, path, delimiter, db):
        try:
            print(delimiter)
            path=path[0]
            self.logger.info(f"Reading data from {db} CSV File")
            df = pd.read_csv(path, sep=delimiter)
            df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
            columns = df.columns.values.tolist()
            self.reporter.addRow(
                f"Parsing {db} file", "Parsing file is successful", status.PASS)
            return df, columns
        except Exception as e:
            raise Exception(f"Could not read file, {e}")

    def dbOperations(self, cred, db):
        try:
            log = f"----Connecting to {db}DB----"
            self.logger.info(log)
            myDB = self.connectingDB(db, cred)
        except Exception as e:
            print("DB Operations..........................................")
            self.reporter.addRow(
                f"Connection to {db}db: ", "Exception occurred", status.ERR)
            self.logger.error(str(e))
            raise Exception(e)
        
        try:
            db_1, columns = self.executeQuery(db, myDB)
        except Exception as e:
            self.logger.error(str(e))
            self.reporter.addRow(
                f"Executing {db} SQL", "Exception Occurred", status.ERR)
            raise e
        return db_1,columns
    
    def fetch_data_with_retries(self, cred, db):
        retry_limit = 3  # Maximum number of connection retry attempts
        retry_delay = 10  # Delay between retry attempts in seconds
        retries = 0
        while retries < retry_limit:
            try:
                return self.dbOperations(cred, db)
            except Exception as e:
                print(f"Connection failed: {str(e)}")
                print("Retrying...")
                print("fetch_data_with_retries...........................")
                time.sleep(retry_delay)
                retries += 1

        print("Maximum number of retries exceeded. Failed to fetch data.")
        return None, None

    
    def executeQuery(self,db,myDB):
        try:
            self.logger.info(f"----Executing the {db}SQL----")
            # checking whether it is mongodb object by checking mongoclient in starting of the object
            x = re.search("^MongoClient.*", str(myDB))
            if x:
                try:
                    db_name = f'{db}_DB'
                    db_details = ast.literal_eval(self.configData[db_name])
                    sql = f'{db}_SQL'
                    conn = myDB[db_details['database']]
                    list_of_collections = conn.list_collection_names()
                    if db_details['collection'] not in list_of_collections:
                        raise Exception
                    conn = conn[db_details['collection']]
                except Exception:
                    self.reporter.addRow("Finding database and collection","Not found in source_db or target_db",status.ERR)
                    raise Exception("Database and collection not found in source_db or target_db")
                try:
                    query = json.loads(self.configData[sql])
                except Exception:
                    self.reporter.addRow("Fetching query","Not present in json format",status.ERR)
                    raise Exception("Mongo query not present in json format")
                results = list(conn.find(query))
                db_1 = pd.DataFrame(results)
                columns = list(db_1.columns)
            else:
                # try:
                self.logger.warn("Inside sql execution left mongo db part")
                db_name = f'{db}_DB'
                myCursor = myDB.cursor()
                sql = f"{db}_SQL"
                self.logger.warn(f"SQL: {self.configData[sql]} {sql} {db}")
                self.logger.warn(f"Executing SQL:\n{self.configData[sql]}")
                statement_timeout = 0
                # myDB.cursor().execute(f"ALTER SESSION SET STATEMENT_TIMEOUT_IN_SECONDS = {statement_timeout}")
                # myDB.cursor().execute(f"alter warehouse COMPUTE_WH set statement_timeout_in_seconds = {statement_timeout}")
                myCursor.execute(self.configData[sql])
                self.logger.warn(f"Executed SQL:\n{self.configData[sql]}")
                self.reporter.addRow(
                    f"Executing {db} sql", f"{self.configData[sql]}<br>{db} sql executed successfull", status.PASS)
                columns = [i[0] for i in myCursor.description]
                results = myCursor.fetchall()
                db_1 = pd.DataFrame(results,columns=columns)
        except Exception as e:
            self.logger.error(str(e))
            # self.reporter.addMisc(
            #         "REASON OF FAILURE", get_reason_of_failure(traceback.format_exc(), e))

            self.reporter.addRow(
                f"Executing {db} sql", "Exception occurred", status.ERR)
            raise e
        
        myDB.close()
        return db_1, columns

    def connectingDB(self, dbType, cred):
        
        db = f'{dbType}_DB'
        is_dict = True
        try:
            db_details = ast.literal_eval(self.configData[db])
        except Exception:
            is_dict = False
        if is_dict == True:
            db_type = db_details.get('type')
        else:
            db_type = self.configData[db]
                    
        if db_type.lower() == 'custom':
            conn = f"{dbType}_CONN"
            db = self.configData[conn]
            myDB = ast.literal_eval(db)
            # this is just to check whether connection is established or not
            x = re.search("^MongoClient.*", str(myDB))
            if x:
                myDB.server_info()
            else:
                dbCursor = myDB.cursor()
            self.reporter.addRow(
                f"Connection to {dbType}db:", f"Connection to {dbType}db is successfull", status.PASS)
        else:
            dv = Databases()
            lib, connect = Databases.getConnectionString(
                dv, db_type)
            lib = importlib.import_module(lib)
            connection = getattr(lib, connect)
            myDB = connection(**cred)
            x = re.search("^MongoClient.*", str(myDB))
            if x:
                myDB.server_info()
            connDetails = self.getHostDetails(cred)
            self.reporter.addRow(
                f"Connection to {dbType}db: {connDetails}", f"Connection to {dbType}db is successful", status.PASS)
        return myDB

    def getHostDetails(self, details):

        li = list(details.keys())
        for i in li:
            if re.search("user", i) != None:
                del details[i]
                continue
            if re.search("password", i) != None:
                del details[i]
        return details


import pandas as pd
import traceback
import json
from gempyp.libs.common import dateTimeEncoder, findDuration
from datetime import datetime, timezone
import numpy as np
from gempyp.engine import dataUpload
import logging
import os
from gempyp.config import DefaultSettings
from cryptography.fernet import Fernet
from gempyp.config.DefaultSettings import encrypt_key, checkUrl

class TestData:
    def __init__(self):
        """
        declairing some attribute the are used in testcaseDetails in report and return a object that is used to update data
        """
        self.testcaseDetailColumn = [
            "tc_run_id",
            "start_time",
            "end_time",
            "name",
            "category",
            "log_file",
            "status",
            "steps",
            "machine",
            "result_file",
            "product_type",
            "ignore",
            "meta_data",
            "user_defined_data",
            "base_user",
            "invoke_user",
            "run_type",
            "run_mode"
        ]
        self.misc_detail_column = ["run_id", "key", "value", "table_type"]

        # can have anyamount of columns
        # this should always have one row so it can be made a dict or something instad of a dataframe
        self.suite_detail = pd.DataFrame()

        self.testcase_details = pd.DataFrame(columns=self.testcaseDetailColumn)
        self.misc_details = pd.DataFrame(columns=self.misc_detail_column)

    def toSuiteJson(self):
        """
        converts the dataframe to Json
        used in uploadsuitedata (run() method in engine.py)
        """
        if self.suite_detail.empty:
            return {}
        
        self.suite_detail = self.suite_detail.replace(np.nan, '-', regex=True)
        data = self.suite_detail.to_dict(orient="records")[0]
        misc_data = self.misc_details[
            self.misc_details["table_type"].str.upper() == "SUITE"
        ]
        # the above misc data is not being used

        data["meta_data"] = data["meta_data"]
        return json.dumps(data, cls=dateTimeEncoder)

    def totestcaseJson(self, tc_run_id, s_run_id):
        """
        returns the json for testcasedata
        used in update_df method of engine.py
        """

        self.testcase_details = self.testcase_details.replace(np.nan, '-', regex=True)
        test_data = self.testcase_details.loc[
            self.testcase_details["tc_run_id"].str.upper() == tc_run_id
        ]
        if test_data.empty:
            return {}
        test_data = test_data.to_dict(orient="records")[0]
        misc_data = self.misc_details.loc[self.misc_details["run_id"].str.upper() == tc_run_id]
        misc_data = misc_data.to_dict(orient="records")
        test_status = {}
        for step in test_data["steps"]:
            key = step.get("status")
            if test_status.get(key, None) is not None:
                test_status[key] = test_status.get(key) + 1
            else:
                test_status[key] = 1
        test_status["TOTAL"] = sum(test_status.values())
        prio_list = ['TOTAL', 'PASS', 'FAIL']
        sorted_dict = {}
        for key in prio_list:
            if key in test_status:
                sorted_dict[key] = test_status[key]
                test_status.pop(key)
            elif key == "PASS" or key == 'FAIL':
                sorted_dict[key] = 0
        sorted_dict.update(test_status)
        # test_data["duration"] = findDuration(test_data["start_time"], test_data["end_time"])

        test_data["user_defined_data"] = dict()
        # test_data["duration"]="{0:.{1}f} sec(s)".format((test_data["end_time"]-test_data["start_time"]).total_seconds(),2)
        """ Adding misc data to user_defined_data column for each testcase
        Here misc data is only for one testcase.
        {"key1": "value1", "key2": "value2"...}"""
        if len(misc_data) > 0:
            for miscs in misc_data:
                # logging.info("--- misc key", miscs.get("key", None))
                key = str(miscs["key"])
                val = str(miscs["value"])
                test_data["user_defined_data"][key] = val

        meta_data = [
            {
                "TESTCASE NAME": test_data["name"], 
                "SERVICE PROJECT": "None", 
                "DATE OF EXECUTION": {"value": datetime.now(timezone.utc), "type": "date"},
                # "RESPONSE TIME":"{0:.{1}f} sec(s)".format((test_data["end_time"]-test_data["start_time"]).total_seconds(),2)
            }, 
            {
                "EXECUTION STARTED ON": {"value": test_data["start_time"], "type": "datetime"},
                "EXECUTION ENDED ON": {"value": test_data["end_time"], "type": "datetime"}, 
                "EXECUTION DURATION": findDuration(test_data["start_time"], test_data["end_time"])
            }, 
            sorted_dict
            # {"response_time":"{0:.{1}f} sec(s)".format((test_data["end_time"]-test_data["start_time"]).total_seconds(),2)}
            ]

        test_data["meta_data"] = meta_data
        test_data["s_run_id"] = s_run_id
        test_data["user_defined_data"]["duration"]="{0:.{1}f} sec(s)".format((test_data["end_time"]-test_data["start_time"]).total_seconds(),2)
        
        return json.dumps(test_data, cls=dateTimeEncoder)

    def _validate(self):
        """
        used by fromJSON to validate the json data
        """
        pass

    def getJSONData(self):
        """
        provide the report json to makereport() method of engine file
        """
        suite_report = {}
        self.suite_detail = self.suite_detail.replace(np.nan, "-", regex=True)
        suite_dict = self.suite_detail.to_dict(orient="records")[0]
        testcase_dict = self.testcase_details.to_dict(orient="records")
        misc_dict = self.misc_details.to_dict(orient="records")
        try:

            # converting testcase_dict to dict for easy parsing
            test_dict = {d['tc_run_id']: d for d in testcase_dict}
            key = list(test_dict.keys())
            key = key[0]
            test_data = test_dict[key]
            # Don't have idea about this code needed to be restructures  30-07-2025 Sachin

            # test_dict[key] = test_data
            # for each_misc in misc_dict:
            #     test_dict[each_misc['run_id']][each_misc["key"]] = each_misc["value"]

            # converting dict back to list
            testcase_dict = [test_dict[tc_run_id] for tc_run_id in test_dict.keys()]
        except Exception as e:
            traceback.print_exc()

        # print(testcase_dict, "++++++++++++++++++++++++++++++++++++++++")
        for i in range(len(testcase_dict)):
            for key in ["meta_data", "base_user", "invoke_user", "user_defined_data","LOG_FILE"]:
                if(key in testcase_dict[i].keys()):
                    testcase_dict[i].pop(key)
        suite_dict["testcase_details"] = testcase_dict
        testcase_counts = self.getTestcaseCounts()

        prio_list = ['total', 'PASS', 'FAIL']  # to be optimized
        sorted_dict = {}
        for key in prio_list:
            if key in testcase_counts :
                sorted_dict[key] = testcase_counts[key]
                testcase_counts.pop(key)
            elif key == "PASS" or key == 'FAIL':
                sorted_dict[key] = 0
        sorted_dict.update(testcase_counts)
        testcase_counts = sorted_dict
        suite_dict["testcase_info"] = testcase_counts
        suite_report["suits_details"] = suite_dict
        suite_report["report_product"] = "GEMPYP"       
        return json.dumps(suite_report, cls=dateTimeEncoder)

    def getTestcaseCounts(self):
        """
        return the no. of testcases
        """
        group = self.testcase_details.groupby(["status"]).size()
        group = group.to_dict()
        total = 0
        for i in group:
            total += group[i]
        group["total"] = total

        return group

    def validateSrunidInDB(self,jewel_user,s_run_id,username,bridgetoken):
        if jewel_user:
            # if "RUN_ID" in params:
            if s_run_id:
                logging.info("************Trying to check If s_run_id is present in DB*****************")
                # response =  dataUpload.checkingData(s_run_id, params["BRIDGE_TOKEN"], params["USERNAME"])
                if DefaultSettings.apiSuccess:
                    if not dataUpload.checkingData(s_run_id,bridgetoken, username):
                        logging.info("************s_run_id not present in DB Trying to call Post*****************")
                        dataUpload.sendSuiteData((self.toSuiteJson()), bridgetoken,username)
                    else:
                        logging.info("---------s_run_id already present --------------")
                        dataUpload.sendSuiteData((self.toSuiteJson()), bridgetoken, username,mode="PUT")
            else:
                if DefaultSettings.apiSuccess:
                    dataUpload.sendSuiteData((self.toSuiteJson()), bridgetoken, username)
                    ### first try to rerun the data
                    self.retryUploadSuiteData(bridgetoken,username)

    
    def retryUploadSuiteData(self,bridgetoken,username):
            # if dataUpload.suite_uploaded == False:
            if not dataUpload.suite_uploaded:
                        logging.info("------Retrying to Upload Suite Data------")
                        print(self.toSuiteJson())
                        dataUpload.sendSuiteData((self.toSuiteJson()), bridgetoken, username)


    def retryUploadTestcases(self,s_run_id,bridgetoken,username,output_folder):
        jewel = ''
        unuploaded_path = ""
        failed_Utestcases=0
        # if dataUpload.suite_uploaded == True:
        if dataUpload.suite_uploaded:
            jewelLink = DefaultSettings.getUrls('jewel-url')
            jewel = f'{jewelLink}/#/autolytics/execution-report?s_run_id={s_run_id}&p_id={DefaultSettings.project_id}'
        if len(dataUpload.not_uploaded) != 0 and dataUpload.suite_uploaded:
            logging.info("------Trying again to Upload Testcase------")
            for testcase in dataUpload.not_uploaded:
                dataUpload.sendTestcaseData(testcase,bridgetoken, username)
            # if dataUpload.flag:
                # logging.warning("Testcase may be present with same tc_run_id in database")
            # unuploaded_path=self.unuploadedFile(output_folder,dataUpload.not_uploaded,"Unploaded_testCases.json")
            # failed_Utestcases = len(dataUpload.not_uploaded) 
            ### Creating file for unuploaded testcases
            # if len(dataUpload.not_uploaded) != 0:
            #     if dataUpload.flag == True:
            #         logging.warning("Testcase may be present with same tc_run_id in database")
            #     unuploaded_path=self.unuploadedFile(output_folder,dataUpload.not_uploaded,"Unploaded_testCases.json")
        # return jewel,failed_Utestcases,unuploaded_path
        return jewel,failed_Utestcases


    def WriteSuiteFile(self,base_url,output_folder,username,bridge_token):
            if not base_url:
                logging.warning("Maybe username or bridgetoken is missing or wrong thus data is not uploaded in db.")
            dataUpload.suite_data.append(self.toSuiteJson())
            unuploaded_dict = {}
            unuploaded_dict["suite_data"] = dataUpload.suite_data
            unuploaded_dict["testcases"] = dataUpload.not_uploaded
            # unuploaded_dict["urls"] = DefaultSettings.urls['data']
            unuploaded_dict["base_url"] = checkUrl(base_url)
            unuploaded_dict["user_name"] = username
            unuploaded_dict["bridge_token"] = bridge_token
            unuploaded_path=self.unuploadedFile(output_folder,unuploaded_dict,"Unuploaded_data.json")
            return unuploaded_path


    ### Function to create unuploadedFile
    def unuploadedFile(self,output_folder,data,fileName):
        unuploaded_path = ""
        # listToStr = ',\n'.join(map(str, data))
        unuploaded_path = os.path.join(output_folder, fileName)
        data = str(data)
        fernet = Fernet(encrypt_key)
        data = fernet.encrypt(data.encode())
        with open(unuploaded_path,'wb') as w:
            w.write(data)
        return unuploaded_path
        


   



import json
import logging
from time import time
import traceback
from datetime import datetime
from getpass import getuser
import warnings
from cryptography.fernet import Fernet
import requests
import requests.auth
from requests_ntlm import HttpNtlmAuth
import os


class Api:
    def __init__(self):
        pass

    def execute(self, request):
        encrypt_key = "Y4irRsiBmyGMBie5gAZ8va3IOHVOYZFxC5L1-jNydZk="
        result = Response()
        # if not request.file:
        header_dict = {key.upper(): value.upper() for key, value in request.headers.items()}
        if header_dict.get("CONTENT-TYPE", "") == "APPLICATION/JSON":
                try:
                    if not isinstance(request.body, str):
                        request.body = json.dumps(request.body)
                except Exception as e:
                    logging.info("JSON object can not be serialized")
                    logging.info(str(e))
        else:
                try:
                    newFile=[]
                    newBody={}
                    for key,value in request.body.items():
                        newFileTuple=tuple()
                        if(isinstance(value,str) and os.path.isfile(value)):
                            name = value.split('/')[-1]
                            newFileTuple+=(key,(name,open(value,'rb')))
                            newFile.append(newFileTuple)
                        else:
                            if(isinstance(value,dict)):
                                newBody[key]=str(json.dumps(value))
                            else:
                                newBody[key]=value
                    request.file=newFile
                    request.body=newBody
                    # request.body = self.convert_quotes_boolean(request.body)
                    pass
                except Exception as e:
                    traceback.print_exc()
                    logging.info("JSON object can not be serialized")
                    logging.info(str(e))
        auth = None
        try:
                if request.auth.upper() == "PASSWORD":
                    password = self.decrypt_(request.credentials.get("password", ""), encrypt_key)
                    auth = HttpNtlmAuth(request.credentials.get("username", " "), password)
        except Exception as e:
                logging.info("Error occured while creating the auth object- " + str(e))
                auth = None
        try:

                start_time = end_time = datetime.now()
                obj = Response()
                resp, start_time, end_time = self.make_request(request)

                obj.status_code = resp.status_code
                obj.response_body = resp.text
                obj.response_headers = resp.headers
                obj.response_time = resp.elapsed.total_seconds()

                

                if obj.response_time == 0:
                    elapsed_time = end_time - start_time
                    obj.response_time = elapsed_time.total_seconds()
                logging.info(f"URL: {request.api}")
                logging.info(f"Time elapsed: {obj.response_time} secs")
                
                result = obj
                
        except Exception as e:
                traceback.print_exc()
                logging.info(str(e))
                result = obj
            
            # add retries if required
        if request.retries > 0:
                if result.status_code not in request.expected_code_list:
                    request.retries -= 1
                    logging.info("retrying...........")
                    time.sleep(1)
                    return self.execute_api(request)
        
        return result

    def make_request(self,request):
        METHODS = {
            "POST": requests.post,
            "GET": requests.get,
            "PUT": requests.put,
            "PATCH": requests.patch,
            "DELETE": requests.delete,
        }
        if not request.api or not request.method:
            raise Exception("Api and method cannot be empty or NULL")

        # Convert method to uppercase for consistency
        method = request.method.upper()

        if method not in METHODS:
            raise Exception("Invalid method")

        # Define headers, files, data, and authentication based on the request
        auth = None
        if request.auth and request.auth.upper() == "PASSWORD":
            auth = (request.username, request.password)

        headers = request.headers
        files = request.file
        data = request.body
        verify = request.SSLVerify
        timeout = request.timeout

        # Make the request and calculate the time taken
        start_time = datetime.now()
        try:
            resp = METHODS[method](
            request.api,
            headers=headers,
            files=files,
            data=data,
            verify=verify,
            auth=auth,
            timeout=timeout
        )
            end_time = datetime.now()
            return resp, start_time, end_time   
        except Exception as e:
            end_time = datetime.now()
            logging.info("Error while making {http_method} to the {api}".format(http_method=method, api=request.api))
            traceback.format_exception_only()
            return resp, start_time, end_time   
                  



    def encrypt_(self, password, key):
        try:
            fernet_ = Fernet(key)
            # supress user warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                enc_pass = fernet_.encrypt(password)
        except Exception as e:
            enc_pass = None
            traceback.print_exc()
            logging.info(e)
        return enc_pass

    def decrypt_(self, enc_pass, key):
        try:
            fernet_ = Fernet(key)
            # supress user warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                enc_pass = enc_pass.encode('utf-8')
                dec_pass = fernet_.decrypt(enc_pass)
                dec_pass = dec_pass.decode('utf-8')
        except Exception as e:
            dec_pass = None
            traceback.print_exc()
            logging.info(e)
        return dec_pass

         


class Request:
    def __init__(self):
        self.api = ""
        self.method = "method"
        self.headers = {}
        self.body = {}
        self.file = ""
        self.SSLVerify = False
        self.credentials = {}
        self.timeout = 330000
        self.auth = ""
        self.expected_code_list = []
        self.retries = 0


class Response:
    def __init__(self, body="", code=504, response_time=0, headers={}):
        self.response_body = body
        self.status_code = code
        self.response_time = response_time
        self.response_headers = headers



from traceback import print_exc, print_stack
import traceback
from gempyp.libs.enums.status import status
import logging


##   check if actual value is string or not

def compareTo(obj, key, value, key_val_dict, tolerance=0.1, isLegacyPresent = False, isLegacyResponse = False):
    """checks for equality of actual value and expected value. 
    OPERATOR -"to"
    """
    actual_value = key_val_dict.get(key, key)
    if isinstance(actual_value, (int, float)):
        actual_value = str(actual_value)
        exp_value = str(value.strip('"').strip("'"))
    else:
        actual_value = str(actual_value).lower()
        exp_value = value.strip('"').lower()
        if actual_value == value:
            exp_value = value 
        
    if actual_value == exp_value:
        if isLegacyPresent:
            if isLegacyResponse:
                obj.addRow(f"Running validation for {key}",f"Operator = To",status.PASS, CURRENT_API = "-",LEGACY_API = f"Expected:-- {str(exp_value)} <br> Actual:-- {str(actual_value)} <br>condition satisfied" )
            else:
                obj.addRow(f"Running validation for {key}",f"Operator = To",status.PASS, CURRENT_API=f"Expected:-- {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied",LEGACY_API="-" )
        else:
            obj.addRow(f"Running validation for {key}", f"Expected:-- {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied", status.PASS)
    else:
        if isLegacyPresent:
            if isLegacyResponse:
                obj.addRow(f"Running validation for {key}",f"Operator = To",status.FAIL, CURRENT_API="-",LEGACY_API = f"Expected:-- {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition not satisfied" )
            else:
                obj.addRow(f"Running validation for {key}",f"Operator = To",status.FAIL, CURRENT_API=f"Expected:-- {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition not satisfied",LEGACY_API="-" )
        else:
            obj.addRow(f"Running validation for {key}", f"Expected:-- {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition not satisfied", status.FAIL)
        obj.addMisc("REASON OF FAILURE", "Mismatches found during Assertion")


    return obj


def compareToResp(obj,key,value, key_val_dict, key_val_dict_legacy, keyCheckResult, tolerance):
    try:
        actual_value = key_val_dict.get(key,key)
        required_value = key_val_dict_legacy[value]
        if keyCheckResult == "FOUND":
            if actual_value == required_value:
                if 'legacy' in key:
                    obj.addRow(f"Comparing  |  Respective values of {key} and {value}",f"Check if values are equal.",status.PASS, CURRENT_API=f"Value for {value} is {key_val_dict_legacy[value]}", LEGACY_API=f"Value for {key} is {key_val_dict[key]}")
                else:
                    obj.addRow(f"Comparing  |  Respective values of {key} and {value}",f"Check if values are equal",status.PASS, CURRENT_API=f"Value for {key} is {key_val_dict[key]}", LEGACY_API=f"Value for {value} is {key_val_dict_legacy[value]}")
            else:
                if 'legacy' in key:
                    obj.addRow(f"Comparing  |  Respective values of {key} and {value}",f"Check if values are equal",status.FAIL, CURRENT_API=f"Value for {value} is {key_val_dict_legacy[value]}", LEGACY_API=f"Value for {key} is {key_val_dict[key]}")
                else:
                    obj.addRow(f"Comparing  |  Respective values of {key} and {value}",f"Check if values are equal",status.FAIL, CURRENT_API=f"Value for {key} is {key_val_dict[key]}", LEGACY_API=f"Value for {value} is {key_val_dict_legacy[value]}")
    except:
        traceback.print_exc()

def compareNotToResp(obj,key,value, key_val_dict, key_val_dict_legacy, keyCheckResult, tolerance):
    try:
        actual_value = key_val_dict.get(key,key)
        required_value = key_val_dict_legacy[value]
        if keyCheckResult == 'FOUND':
            if actual_value != required_value:
                if 'legacy' in key:
                    obj.addRow(f"Running Assertion on comparing the respective values of {key} and {value}",f"Check if values are not equal.",status.PASS, CURRENT_API=f"Value for {value} is {key_val_dict_legacy[value]}", LEGACY_API=f"Value for {key} is {key_val_dict[key]}")
                else:
                    obj.addRow(f"Running Assertion on comparing the respective values of {key} and {value}",f"Check if values are not equal.",status.PASS, CURRENT_API=f"Value for {key} is {key_val_dict[key]}", LEGACY_API=f"Value for {value} is {key_val_dict_legacy[value]}")
            else:
                if 'legacy' in key:
                    obj.addRow(f"Running Assertion on comparing the respective values of {key} and {value}",f"Check if values are not equal.",status.FAIL, CURRENT_API=f"Value for {value} is {key_val_dict_legacy[value]}", LEGACY_API=f"Value for {key} is {key_val_dict[key]}")
                else:
                    obj.addRow(f"Running Assertion on comparing the respective values of {key} and {value}",f"Check if values are not equal.",status.FAIL, CURRENT_API=f"Value for {key} is {key_val_dict[key]}", LEGACY_API="")
                    obj.addMisc("REASON OF FAILURE", "Mismatches found during Assertion")
    except:
        traceback.print_exc()

            
    

def compareNotTo(obj, key, value, key_val_dict, tolerance=0.1, isLegacyPresent = False, isLegacyResponse = False):
    """checks for inequality of actual value and expected value. 
    OPERATOR -"notto, not_to"
    """
    
    actual_value = key_val_dict.get(key, key)
    if isinstance(actual_value, (int, float)):
        actual_value = str(actual_value)
        exp_value = str(value.strip('"').strip("'"))
    else:
        actual_value = str(actual_value).lower()
        exp_value = value.strip('"').strip("'").lower()

    if actual_value != exp_value:
        if isLegacyPresent:
            if isLegacyResponse:
                obj.addRow(f"Running validation for {key}",f"Operator = Not To",status.PASS, CURRENT_API = "-",LEGACY_API = f"Expected:-- != {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied" )
            else:
                obj.addRow(f"Running validation for {key}",f"Operator = Not To",status.PASS, CURRENT_API=f"Expected:-- {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied",LEGACY_API="-" )
        else:
            obj.addRow(f"Running validation for {key}", f"Expected:-- != {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied", status.PASS)
    else:
        if isLegacyPresent:
            if isLegacyResponse:
                obj.addRow(f"Running validation for {key}",f"Operator = Not To",status.FAIL, CURRENT_API="-",LEGACY_API = f"Expected:-- != {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition not satisfied" )
            else:
                obj.addRow(f"Running validation for {key}",f"Operator = Not To",status.FAIL, CURRENT_API=f"Expected:-- != {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition not satisfied",LEGACY_API="-" )
        else:
            obj.addRow(f"Running validation for {key}", f"Expected:-- != {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition not satisfied", status.FAIL)
            obj.addMisc("REASON OF FAILURE", "Mismatches found during Assertion")
    
    return obj


def compareIn(obj, key, value, key_val_dict, tolerance=0.1, isLegacyPresent = False, isLegacyResponse = False):
    """Checks whether the actual value is in the list of expecteds value or not
    OPERATOR - "in"
    """
    actual_value = key_val_dict.get(key, key)

    exp_value = [i.strip('"').strip("'").lower() for i in list(value.strip("[]").split(","))]
    if isinstance(actual_value, (int, float)):
        actual_value = actual_value
        try:
            exp_value = [float(i) for i in exp_value]
        except Exception as e:
            logging.info(str(e)) 
    else:
        actual_value = str(actual_value).lower()
    if actual_value in exp_value:
        if isLegacyPresent:
            if isLegacyResponse:
                obj.addRow(f"Running validation for {key}" ,f"Operator = in", status.PASS, CURRENT_API = "-",LEGACY_API = f"Expected:-- in list {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied")
            else:
                obj.addRow(f"Running validation for {key}" ,f"Operator = in", status.PASS, CURRENT_API = f"Expected:-- in list {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied",LEGACY_API = "-")
        else:
            obj.addRow(f"Running validation for {key}", f"Expected:-- in list {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied", status.PASS)
    else:
        if isLegacyPresent:
            if isLegacyResponse:
                obj.addRow(f"Running validation for {key}",f"Operator = in", status.FAIL,CURRENT_API="-", LEGACY_API=f"Expected:-- in list {str(exp_value)} <br> Actual:-- {str(actual_value)}condition not satisfied")
            else:
                obj.addRow(f"Running validation for {key}",f"Operator = in", status.FAIL,CURRENT_API=f"Expected:-- in list {str(exp_value)} <br> Actual:-- {str(actual_value)}condition not satisfied", LEGACY_API="-")
        else:
            obj.addRow(f"Running validation for {key}", f"Expected:-- in list {str(exp_value)} <br> Actual:-- {str(actual_value)}condition not satisfied", status.FAIL)
            obj.addMisc("REASON OF FAILURE", "Mismatches found during Assertion")
    

    return obj


def compareNotIn(obj, key, value, key_val_dict, tolerance=0.1, isLegacyPresent = False, isLegacyResponse = False):
    """Checks whether the actual value is absent from the not expected values or not.
    OPERATOR - "notin, not_in"
    """
    actual_value = key_val_dict.get(key, key)

    exp_value = [i.strip('"').strip("'").lower() for i in list(value.strip("[]").split(","))]

    if isinstance(actual_value, (int, float)):
        actual_value = actual_value
        try:
            exp_value = [float(i) for i in exp_value]
        except Exception as e:
            logging.info(str(e))
    else:
        actual_value = str(actual_value).lower()
        exp_value = value.strip('"').strip("'").lower()
    if actual_value not in exp_value:
        if isLegacyPresent:
            if isLegacyResponse:
                obj.addRow(f"Running validation for {key}",f"Operator = not in",status.PASS,CURRENT_API="-", LEGACY_API=f"Expected:-- not in list {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied")
            else:
                obj.addRow(f"Running validation for {key}",f"Operator = not in",status.PASS,CURRENT_API=f"Expected:-- not in list {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied", LEGACY_API="-")
        else:
            obj.addRow(f"Running validation for {key}", f"Expected:-- not in list {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition satisfied", status.PASS)
    else:
        if isLegacyPresent:
            if isLegacyResponse:
                obj.addRow(f"Running validation for {key}",f"Operator = not in",status.FAIL,CURRENT_API="-", LEGACY_API=f"Expected:-- not in list {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition not satisfied")
            else:
                obj.addRow(f"Running validation for {key}",f"Operator = not in",status.FAIL,CURRENT_API=f"Expected:-- not in list {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition not satisfied", LEGACY_API="-")
        else:
            obj.addRow(f"Running validation for {key}", f"Expected:-- not in list {str(exp_value)} <br> Actual:-- {str(actual_value)} <br> condition not satisfied", status.FAIL)
            obj.addMisc("REASON OF FAILURE", "Mismatches found during Assertion")
        
    return obj

def compareContains(obj, key, value, key_val_dict, tolerance=0.1, isLegacyPresent = False, isLegacyResponse = False):
    """Checks whether the actual value contains(substring) expected value or not
    OPERATOR - "contains"
    """
    actual_value = key_val_dict.get(key, key)
    if value.lower().strip("'").strip('"') in str(actual_value).strip("'").strip('"').lower():
        if isLegacyPresent:
            if isLegacyResponse:
                obj.addRow(f"Running validation for {key}",f"Operator = contains",status.PASS, CURRENT_API="-", LEGACY_API=f"Expected:-- contains {value} <br> Actual:-- {str(actual_value)} <br> condition satisfied")
            else:
                obj.addRow(f"Running validation for {key}",f"Operator = contains",status.PASS, CURRENT_API=f"Expected:-- contains {value} <br> Actual:-- {str(actual_value)} <br> condition satisfied", LEGACY_API="-")
        else:
            obj.addRow(f"Running validation for {key}", f"Expected:-- contains {value} <br> Actual:-- {str(actual_value)} <br> condition satisfied", status.PASS)
    else:
        if isLegacyPresent:
            if isLegacyResponse:
                obj.addRow(f"Running validation for {key}",f"Operator = contains", status.FAIL, CURRENT_API="-", LEGACY_API=f"Expected:-- contains {value} <br> Actual:-- {str(actual_value)} <br> condition not satisfied")
            else:
                obj.addRow(f"Running validation for {key}",f"Operator = contains", status.FAIL, CURRENT_API=f"Expected:-- contains {value} <br> Actual:-- {str(actual_value)} <br> condition not satisfied", LEGACY_API="-")
        else:
            obj.addRow(f"Running validation for {key}", f"Expected:-- contains {value} <br> Actual:-- {str(actual_value)} <br> condition not satisfied", status.FAIL)
            obj.addMisc("REASON OF FAILURE", "Mismatches found during Assertion")
        
    return obj


def noOperator(obj):
    logging.info("operator not supported")
    return obj
